from typing import Optional, Tuple

from textx import metamodel_from_str

from strictdoc.backend.sdoc.grammar.grammar_builder import SDocGrammarBuilder
from strictdoc.backend.sdoc.models.anchor import Anchor
from strictdoc.backend.sdoc.models.free_text import FreeTextContainer
from strictdoc.backend.sdoc.models.inline_link import InlineLink
from strictdoc.backend.sdoc.processor import ParseContext
from strictdoc.helpers.textx import drop_textx_meta


class SDFreeTextReader:
    @staticmethod
    def _read(
        input_string: str, file_path: Optional[str] = None
    ) -> Tuple[FreeTextContainer, ParseContext]:
        meta_model = metamodel_from_str(
            SDocGrammarBuilder.create_free_text_grammar(),
            classes=[FreeTextContainer, InlineLink, Anchor],
            use_regexp_group=True,
        )

        parse_context = ParseContext(file_path)

        document = meta_model.model_from_str(input_string, file_name=file_path)
        parse_context.document_reference.set_document(document)

        # HACK:
        # ProcessPoolExecutor doesn't work because of non-picklable parts
        # of textx. The offending fields are stripped down because they
        # are not used anyway.
        drop_textx_meta(document)

        return document, parse_context

    @staticmethod
    def read(
        input_string: str, file_path: Optional[str] = None
    ) -> FreeTextContainer:
        assert isinstance(input_string, str), input_string
        document, _ = SDFreeTextReader._read(input_string, file_path)
        return document



from typing import Any, Dict, List, Optional

from strictdoc.backend.sdoc.document_reference import DocumentReference
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.model import RequirementFieldName
from strictdoc.backend.sdoc.models.node import SDocNode, SDocNodeField


class SDocObjectFactory:
    @staticmethod
    def create_document(title: Optional[str]) -> SDocDocument:
        return SDocDocument(
            mid=None,
            title=title if title else "NONAME",
            config=None,
            view=None,
            grammar=None,
            section_contents=[],
        )

    @staticmethod
    def create_requirement(
        parent: Any,
        node_type: str = "REQUIREMENT",
        uid: Optional[str] = None,
        level: Optional[str] = None,
        title: Optional[str] = None,
        statement: Optional[str] = None,
        statement_multiline: Optional[str] = None,
        rationale: Optional[str] = None,
        rationale_multiline: Optional[str] = None,
        tags: Optional[str] = None,
        comments: Optional[List[str]] = None,
    ) -> SDocNode:
        fields: List[SDocNodeField] = []
        if uid is not None:
            assert isinstance(uid, str) and len(uid) > 0
            fields.append(
                SDocNodeField.create_from_string(
                    parent=None,
                    field_name=RequirementFieldName.UID,
                    field_value=uid,
                    multiline=False,
                )
            )
        if level:
            assert isinstance(level, str) and len(level) > 0
            fields.append(
                SDocNodeField.create_from_string(
                    parent=None,
                    field_name=RequirementFieldName.LEVEL,
                    field_value=level,
                    multiline=False,
                )
            )
        if title is not None:
            fields.append(
                SDocNodeField.create_from_string(
                    parent=None,
                    field_name=RequirementFieldName.TITLE,
                    field_value=title,
                    multiline=False,
                )
            )
        if statement:
            fields.append(
                SDocNodeField.create_from_string(
                    parent=None,
                    field_name=RequirementFieldName.STATEMENT,
                    field_value=statement,
                    multiline=False,
                )
            )
        if statement_multiline is not None:
            assert isinstance(statement_multiline, str), (
                f"{statement_multiline}"
            )
            fields.append(
                SDocNodeField.create_from_string(
                    parent=None,
                    field_name=RequirementFieldName.STATEMENT,
                    field_value=statement_multiline,
                    multiline=True,
                )
            )
        if rationale:
            fields.append(
                SDocNodeField.create_from_string(
                    parent=None,
                    field_name=RequirementFieldName.RATIONALE,
                    field_value=rationale,
                    multiline=False,
                )
            )
        if rationale_multiline:
            fields.append(
                SDocNodeField.create_from_string(
                    parent=None,
                    field_name=RequirementFieldName.RATIONALE,
                    field_value=rationale_multiline,
                    multiline=True,
                )
            )
        if tags is not None:
            assert isinstance(tags, str), f"{tags}"
            fields.append(
                SDocNodeField.create_from_string(
                    parent=None,
                    field_name=RequirementFieldName.TAGS,
                    field_value=tags,
                    multiline=False,
                )
            )
        if comments is not None:
            assert isinstance(comments, list), f"{comments}"
            for comment in comments:
                assert isinstance(comment, str), f"{comment}"
                fields.append(
                    SDocNodeField.create_from_string(
                        parent=None,
                        field_name=RequirementFieldName.COMMENT,
                        field_value=comment,
                        multiline=True,
                    )
                )
        requirement = SDocNode(
            parent=parent,
            node_type=node_type,
            fields=fields,
            relations=[],
        )
        requirement.ng_document_reference = DocumentReference()
        if isinstance(parent, SDocDocument):
            requirement.ng_document_reference.set_document(parent)
        else:
            requirement.ng_document_reference.set_document(
                parent.get_document()
            )
        return requirement

    @staticmethod
    def create_requirement_from_dict(
        requirement_dict: Dict[str, str], parent: Any, level: int
    ) -> SDocNode:
        assert requirement_dict is not None
        assert parent
        assert isinstance(level, int)
        assert level > 0

        uid = None
        if RequirementFieldName.UID in requirement_dict:
            uid_ = requirement_dict[RequirementFieldName.UID]
            if isinstance(uid_, str):
                uid = uid_

        title = None
        if RequirementFieldName.TITLE in requirement_dict:
            title_ = requirement_dict[RequirementFieldName.TITLE]
            if isinstance(title_, str):
                title = title_

        statement_multiline = None
        if RequirementFieldName.STATEMENT in requirement_dict:
            statement_multiline_ = requirement_dict[
                RequirementFieldName.STATEMENT
            ]
            if isinstance(statement_multiline_, str):
                statement_multiline = statement_multiline_

        rationale_multiline = None
        if RequirementFieldName.RATIONALE in requirement_dict:
            rationale_multiline_ = requirement_dict[
                RequirementFieldName.RATIONALE
            ]
            if isinstance(rationale_multiline_, str):
                rationale_multiline = rationale_multiline_

        requirement = SDocObjectFactory.create_requirement(
            parent=parent,
            node_type="REQUIREMENT",
            uid=uid,
            level=None,
            title=title,
            statement=None,
            statement_multiline=statement_multiline,
            rationale=None,
            rationale_multiline=rationale_multiline,
            tags=None,
            comments=None,
        )

        return requirement



class NodeCreationOrder:
    BEFORE = "before"
    CHILD = "child"
    AFTER = "after"

    @staticmethod
    def is_valid(order: str) -> bool:
        return order in (
            NodeCreationOrder.BEFORE,
            NodeCreationOrder.CHILD,
            NodeCreationOrder.AFTER,
        )



import re
from typing import Callable, List, Match, Optional

REGEX_TRAILING_WHITESPACE_SINGLELINE = re.compile(r"\s{2,}")
REGEX_TRAILING_WHITESPACE_MULTILINE = re.compile(r" +\n")


# WIP: Check if this is used.
def escape(string: str) -> str:
    return string.encode("unicode_escape").decode("utf-8")


# WIP: Check if this is used.
def unescape(string: str) -> str:
    return string.encode("utf-8").decode("unicode_escape")


def sanitize_html_form_field(field: str, multiline: bool) -> str:
    assert isinstance(field, str)
    sanitized_field: str = field.strip()
    if multiline:
        return REGEX_TRAILING_WHITESPACE_MULTILINE.sub("\n", sanitized_field)
    sanitized_field = sanitized_field.replace("\r\n", "")
    return REGEX_TRAILING_WHITESPACE_SINGLELINE.sub(" ", sanitized_field)


def is_uppercase_underscore_string(string: str) -> bool:
    first_word = r"[A-Z][A-Z0-9]*"
    other_word = r"[A-Z0-9]+"
    pattern = re.compile(rf"^{first_word}(_{other_word})*$")
    return pattern.match(string) is not None


def is_safe_alphanumeric_string(string: str) -> bool:
    pattern = re.compile(r"^[\w.]+([/\\][\w.]+?)*$")
    return pattern.match(string) is not None


def create_safe_acronym(string: str) -> str:
    words = re.split(r"[\W+]", string)
    acronym = ""
    for word in words:
        if len(word) == 0:
            continue
        if word[0].isalpha():
            acronym += word[0].upper()
    return acronym


def create_safe_title_string(string: str) -> str:
    return re.sub(r"[^\w0-9]+", "-", string).rstrip("-")


def extract_last_numeric_part(string: str) -> str:
    regex_match = re.match(".*?([0-9]+)$", string)
    if regex_match is not None:
        return regex_match.group(1)
    raise ValueError(f"Cannot extract the numeric part of UID: {string}")


def extract_numeric_uid_prefix_part(string: str) -> Optional[str]:
    regex_match = re.match(r"^([A-Za-z].*?\D)[0-9]+$", string)
    if regex_match is not None:
        return regex_match.group(1)
    return None


def create_safe_requirement_tag_string(string: str) -> str:
    return re.sub(r"[^A-Za-z0-9]+", "_", string).rstrip("_").upper()


def create_safe_document_file_name(string: str) -> str:
    return re.sub(r"[^A-Za-z0-9]+", "_", string).rstrip("_")


def ensure_newline(text: str) -> str:
    return text.rstrip() + "\n"


def interpolate_at_pattern_lazy(
    template: str, value_resolver: Callable[[str], str]
) -> str:
    pattern = r"@(\w+)"

    def replace_variable(match: Match[str]) -> str:
        variable_name = match.group(1)
        return value_resolver(variable_name)

    return re.sub(pattern, replace_variable, template)


def tokenize(text: str) -> List[str]:
    text = text.lower()

    pattern = r"[a-z0-9]+(?:[-_./][a-z0-9]+)*"
    tokens = re.findall(pattern, text)
    return tokens



from collections import defaultdict
from typing import Dict, List, Optional

from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.document_config import (
    DocumentCustomMetadata,
    DocumentCustomMetadataKeyValuePair,
)
from strictdoc.backend.sdoc.models.inline_link import InlineLink
from strictdoc.core.traceability_index import (
    TraceabilityIndex,
)
from strictdoc.core.transforms.validation_error import (
    MultipleValidationError,
)
from strictdoc.export.html.form_objects.document_config_form_object import (
    DocumentConfigFormObject,
)


class UpdateDocumentConfigTransform:
    def __init__(
        self,
        form_object: DocumentConfigFormObject,
        document: SDocDocument,
        traceability_index: TraceabilityIndex,
    ) -> None:
        self.form_object: DocumentConfigFormObject = form_object
        self.document: SDocDocument = document
        self.traceability_index: TraceabilityIndex = traceability_index

    def perform(self) -> None:
        form_object = self.form_object
        document = self.document

        try:
            self.validate(form_object, document)
        except MultipleValidationError:
            raise

        # Update the document.
        document.title = form_object.document_title
        document.config.version = (
            form_object.document_version
            if form_object.document_version is not None
            and len(form_object.document_version) > 0
            else None
        )
        document.config.classification = (
            form_object.document_classification
            if form_object.document_classification is not None
            and len(form_object.document_classification) > 0
            else None
        )
        document.config.requirement_prefix = (
            form_object.document_requirement_prefix
            if form_object.document_requirement_prefix is not None
            and len(form_object.document_requirement_prefix) > 0
            else None
        )
        if len(form_object.custom_metadata_fields):
            entries = [
                DocumentCustomMetadataKeyValuePair(
                    key=field.field_name, value=field.field_value
                )
                for field in form_object.custom_metadata_fields
            ]
            document.config.custom_metadata = DocumentCustomMetadata(
                entries=entries
            )
        else:
            document.config.custom_metadata = None

        self.traceability_index.delete_document(document)

        document.config.uid = (
            form_object.document_uid
            if form_object.document_uid is not None
            and len(form_object.document_uid) > 0
            else None
        )

        self.traceability_index.create_document(document)

    def validate(
        self,
        form_object: DocumentConfigFormObject,
        document: SDocDocument,
    ) -> None:
        errors: Dict[str, List[str]] = defaultdict(list)
        assert isinstance(document, SDocDocument)

        if len(form_object.document_title) == 0:
            errors["TITLE"].append("Document title must not be empty.")

        # Ensure that UID doesn't have any incoming links if it is going to be
        # renamed or removed.
        existing_uid = document.reserved_uid
        new_uid = form_object.document_uid
        if existing_uid is not None:
            if new_uid is None or existing_uid != new_uid:
                existing_incoming_links: Optional[List[InlineLink]] = (
                    self.traceability_index.get_incoming_links(document)
                )
                if (
                    existing_incoming_links is not None
                    and len(existing_incoming_links) > 0
                ):
                    errors["UID"].append(
                        (
                            "Renaming a node UID when the node has "
                            "incoming links is not supported yet. "
                            "Please delete all incoming links first."
                        ),
                    )

        for metadata_field in form_object.custom_metadata_fields:
            if len(metadata_field.field_name) == 0:
                errors[f"METADATA[{metadata_field.field_mid}]"].append(
                    "Key must not be empty."
                )
            if not metadata_field.field_name[0].isalpha():
                errors[f"METADATA[{metadata_field.field_mid}]"].append(
                    "Key must begin with a letter."
                )
            if " " in metadata_field.field_name[0]:
                errors[f"METADATA[{metadata_field.field_mid}]"].append(
                    "Key must not contain spaces."
                )

        if len(errors):
            raise MultipleValidationError(
                "Document form has not passed validation.", errors=errors
            )



"""
@relation(SDOC-SRS-124, scope=file)
"""

from typing import Any, List, Optional

from strictdoc.backend.sdoc_source_code.models.requirement_marker import Req
from strictdoc.helpers.auto_described import auto_described


@auto_described
class LineMarker:
    def __init__(
        self, parent: Any, reqs_objs: List[Req], role: Optional[str] = None
    ) -> None:
        assert isinstance(reqs_objs, list)
        self.parent = parent
        self.reqs_objs: List[Req] = reqs_objs
        self.reqs: List[str] = list(map(lambda req: req.uid, reqs_objs))
        self.role: Optional[str] = (
            role if role is not None and len(role) > 0 else None
        )

        # Line number of the marker in the source code.
        self.ng_source_line_begin: Optional[int] = None
        self.ng_source_column_begin: Optional[int] = None

        self.ng_range_line_begin: Optional[int] = None
        self.ng_range_line_end: Optional[int] = None

        self.ng_is_nodoc: bool = "nosdoc" in self.reqs
        self.begin_or_end = True

    def is_begin(self) -> bool:
        return True

    def is_end(self) -> bool:
        return False

    def is_range_marker(self) -> bool:
        return False

    def is_line_marker(self) -> bool:
        return True

    def is_forward(self) -> bool:
        return False

    def get_description(self) -> Optional[str]:
        return "line"



"""
@relation(SDOC-SRS-34, SDOC-SRS-141, scope=file)
"""

from string import Template

from lark import Lark, ParseTree, UnexpectedToken

from strictdoc.backend.sdoc_source_code.constants import (
    REGEX_REQ,
    RESERVED_KEYWORDS,
)


class GrammarTemplate(Template):
    delimiter = "##"


RELATION_MARKER_START = r"@relation[\(\{]"

GRAMMAR = GrammarTemplate("""
start: ##START

relation_marker: _RELATION_MARKER_START _WS? (relation_node_uid _SEP _WS)+ "scope=" relation_scope ("," _WS "role=" relation_role)? _WS? _BRACE_RIGHT

_RELATION_MARKER_START: /##RELATION_MARKER_START/
relation_node_uid: /##REGEX_REQ/
relation_scope: /file|class|function|line|range_start|range_end/
relation_role: ALPHANUMERIC_WORD

node_field: node_name ":" node_multiline_value
node_name: /(?!(##RESERVED_KEYWORDS))[A-Z_]+/
node_multiline_value: (_WS_INLINE | _NL) (NODE_FIRST_STRING_VALUE _NL) (NODE_STRING_VALUE _NL)*

NODE_FIRST_STRING_VALUE.2: /\\s*[^\n\r]+/x
NODE_STRING_VALUE.2: /(?![ ]*##RELATION_MARKER_START)(?!\\s*[A-Z_]+: )[^\n\r]+/x

_NORMAL_STRING_NO_MARKER_NO_NODE: /(?!\\s*##RELATION_MARKER_START)((?!\\s*[A-Z_]+: )|(##RESERVED_KEYWORDS)).+/

_NORMAL_STRING_NO_MARKER: /(?!\\s*##RELATION_MARKER_START).+/

_BRACE_LEFT: /[\\(\\{{]/
_BRACE_RIGHT: /[\\)\\}}]/

_SEP: ","

_NL : (CR? LF)

_WS : WS
_WS_INLINE : WS_INLINE

ALPHANUMERIC_WORD: /[a-zA-Z0-9_]+/

%import common.WS -> WS
%import common.CR -> CR
%import common.LF -> LF
%import common.WS_INLINE -> WS_INLINE
%import common.NEWLINE -> NEWLINE
""")


class MarkerLexer:
    @staticmethod
    def parse(source_input: str, parse_nodes: bool = False) -> ParseTree:
        if parse_nodes:
            start = "(relation_marker | node_field | _NORMAL_STRING_NO_MARKER_NO_NODE | _WS)*"
        else:
            start = "(relation_marker | _NORMAL_STRING_NO_MARKER | _WS)*"

        grammar = GRAMMAR.substitute(
            RELATION_MARKER_START=RELATION_MARKER_START,
            RESERVED_KEYWORDS=RESERVED_KEYWORDS,
            REGEX_REQ=REGEX_REQ,
            START=start,
        )
        parser: Lark = Lark(
            grammar, parser="lalr", cache=True, propagate_positions=True
        )

        try:
            # FIXME: Without rstrip, there is an edge case where the parser
            #        breaks when resolving conflicts between multiline node
            #        fields and normal strings.
            #        See also test: test_31_single_node_field.
            tree: ParseTree = parser.parse(source_input.rstrip() + "\n")
        except UnexpectedToken as exception_:  # pragma: no cover
            print(  # noqa: T201
                "error: could not parse source comment:\n" + source_input
            )
            raise exception_
        return tree



from dataclasses import dataclass
from typing import Any, List, Optional, Union

from markupsafe import Markup

from strictdoc import __version__
from strictdoc.backend.sdoc.models.anchor import Anchor
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.document_view import NullViewElement
from strictdoc.backend.sdoc.models.node import SDocNode, SDocNodeField
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.backend.sdoc_source_code.models.function_range_marker import (
    FunctionRangeMarker,
)
from strictdoc.backend.sdoc_source_code.models.line_marker import LineMarker
from strictdoc.backend.sdoc_source_code.models.range_marker import (
    ForwardRangeMarker,
    RangeMarker,
)
from strictdoc.backend.sdoc_source_code.models.source_file_info import (
    SourceFileTraceabilityInfo,
)
from strictdoc.core.document_tree import DocumentTree
from strictdoc.core.document_tree_iterator import DocumentTreeIterator
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.source_tree import SourceFile
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.export.html.document_type import DocumentType
from strictdoc.export.html.html_templates import JinjaEnvironment
from strictdoc.export.html.renderers.link_renderer import LinkRenderer
from strictdoc.export.html.renderers.markup_renderer import MarkupRenderer
from strictdoc.helpers.cast import assert_cast


@dataclass
class SourceMarkerTuple:
    ng_range_line_begin: int
    ng_range_line_end: int
    source_line: Markup
    markers: List[
        Union[FunctionRangeMarker, ForwardRangeMarker, LineMarker, RangeMarker]
    ]

    def is_end(self) -> bool:
        return any(map(lambda m_: m_.is_end(), self.markers))

    def is_line_marker(self) -> bool:
        return any(map(lambda m_: isinstance(m_, LineMarker), self.markers))


SourceLineEntry = Union[
    Markup,
    SourceMarkerTuple,
]


@dataclass
class SourceFileViewObject:
    def __init__(
        self,
        *,
        traceability_index: TraceabilityIndex,
        trace_info: SourceFileTraceabilityInfo,
        project_config: ProjectConfig,
        link_renderer: LinkRenderer,
        markup_renderer: MarkupRenderer,
        source_file: SourceFile,
        pygments_styles: Markup,
        pygmented_source_file_lines: List[SourceLineEntry],
        jinja_environment: JinjaEnvironment,
    ):
        self.traceability_index: TraceabilityIndex = traceability_index
        self.trace_info: SourceFileTraceabilityInfo = trace_info
        self.project_config: ProjectConfig = project_config
        self.link_renderer: LinkRenderer = link_renderer
        self.markup_renderer: MarkupRenderer = markup_renderer
        self.source_file: SourceFile = source_file
        self.pygments_styles: Markup = pygments_styles
        self.pygmented_source_file_lines: List[SourceLineEntry] = (
            pygmented_source_file_lines
        )
        self.jinja_environment: JinjaEnvironment = jinja_environment

        self.current_view = NullViewElement()
        self.standalone: bool = False
        self.document_tree_iterator: DocumentTreeIterator = (
            DocumentTreeIterator(
                assert_cast(traceability_index.document_tree, DocumentTree)
            )
        )
        self.is_running_on_server: bool = project_config.is_running_on_server
        self.strictdoc_version = __version__

    def get_document_level(self) -> int:
        return self.source_file.level

    def render_screen(self) -> Markup:
        return self.jinja_environment.render_template_as_markup(
            "screens/source_file_view/index.jinja", view_object=self
        )

    def render_detailed_node_for_banner(self, node_uid: str) -> Markup:
        node: SDocNode = self.traceability_index.get_node_by_uid(node_uid)
        return self.jinja_environment.render_template_as_markup(
            "components/node_content/index_extends_readonly.jinja",
            node=node,
            view_object=self,
            requirement_style="table",
        )

    def render_node_title_for_banner_header(
        self, marker: Union[Any], node_uid: str
    ) -> Markup:
        node: SDocNode = self.traceability_index.get_node_by_uid(node_uid)
        return self.jinja_environment.render_template_as_markup(
            "screens/source_file_view/node_title_for_banner_header.jinja",
            node=node,
            view_object=self,
            role=marker.role,
            is_forward_from_requirements_to_code=marker.is_forward(),
        )

    def render_aside_requirement(
        self,
        node_uid: str,
        range_begin: Optional[str] = None,
        range_end: Optional[str] = None,
    ) -> Markup:
        node: SDocNode = self.traceability_index.get_node_by_uid(node_uid)
        return self.jinja_environment.render_template_as_markup(
            "screens/source_file_view/requirement.jinja",
            requirement=node,
            view_object=self,
            requirement_style="table",
            range_begin=range_begin,
            range_end=range_end,
        )

    def render_node_statement(self, node: SDocNode) -> Markup:
        return self.markup_renderer.render_node_statement(
            DocumentType.DOCUMENT, node
        )

    def render_node_rationale(self, node: SDocNode) -> Markup:
        return self.markup_renderer.render_node_rationale(
            DocumentType.DOCUMENT, node
        )

    def render_node_field(self, node_field: SDocNodeField) -> Markup:
        assert isinstance(node_field, SDocNodeField), node_field
        return self.markup_renderer.render_node_field(
            DocumentType.DOCUMENT, node_field
        )

    def render_url(self, url: str) -> str:
        return self.link_renderer.render_url(url)

    def render_marker_range_link(self, marker: RangeMarker) -> str:
        return self.link_renderer.render_marker_range_link(
            self.source_file.in_doctree_source_file_rel_path_posix,
            self.source_file,
            marker,
        )

    def render_node_link(self, node: SDocNode) -> str:
        assert isinstance(node, SDocNode), node
        return self.link_renderer.render_node_link(
            node, None, DocumentType.DOCUMENT
        )

    def render_static_url(self, url: str) -> str:
        return self.link_renderer.render_static_url(url)

    def render_local_anchor(
        self, node: Union[Anchor, SDocNode, SDocSection, SDocDocument]
    ) -> str:
        return self.link_renderer.render_local_anchor(node)

    def get_source_file_path(self) -> str:
        return self.source_file.in_doctree_source_file_rel_path_posix

    def get_file_stats_lines_total(self) -> str:
        return str(self.trace_info.file_stats.lines_total)

    def get_file_stats_lines_total_non_empty(self) -> str:
        return str(self.trace_info.file_stats.lines_non_empty)

    def get_file_stats_non_empty_lines_covered(self) -> str:
        covered = self.trace_info.ng_lines_covered
        total_non_empty = self.trace_info.file_stats.lines_non_empty
        percentage = (
            (covered / total_non_empty * 100) if total_non_empty > 0 else 0
        )
        return f"{covered} / {total_non_empty} ({percentage:.1f}%)"

    def get_file_stats_functions_total(self) -> str:
        return str(len(self.trace_info.functions))

    def get_file_stats_functions_covered(self) -> str:
        covered = self.trace_info.covered_functions
        total = len(self.trace_info.functions)
        percentage = (covered / total * 100) if total > 0 else 0
        return f"{covered} / {total} ({percentage:.1f}%)"



import os
from pathlib import Path

from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.node import SDocNode
from strictdoc.core.document_iterator import DocumentCachingIterator
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.export.html.renderers.link_renderer import LinkRenderer


class DoxygenGenerator:
    def __init__(self, project_config: ProjectConfig):
        self.project_config: ProjectConfig = project_config

    def export(
        self,
        *,
        traceability_index: TraceabilityIndex,
        path_to_output_dir: str,
    ) -> None:
        Path(path_to_output_dir).mkdir(parents=True, exist_ok=True)
        output_path = os.path.join(path_to_output_dir, "strictdoc.tag")

        link_renderer = LinkRenderer(
            root_path="NOT_RELEVANT", static_path="NOT_RELEVANT"
        )

        def template_node(node_uid: str, path_to_html: str) -> str:
            return f"""\
  <compound kind="file">
    <name>{node_uid}</name>
    <filename>html/{path_to_html}</filename>
  </compound>
"""

        template_all_nodes = ""

        assert traceability_index.document_tree.document_list is not None

        document_: SDocDocument
        for document_ in traceability_index.document_tree.document_list:
            document_iterator = DocumentCachingIterator(document_)

            for node, _ in document_iterator.all_content(
                print_fragments=False,
            ):
                if isinstance(node, SDocNode) and node.reserved_uid is not None:
                    path_to_html = link_renderer.render_node_doxygen_link(node)
                    template_all_nodes += template_node(
                        node.reserved_uid, path_to_html
                    )

        template_xml = f"""\
<?xml version='1.0' encoding='UTF-8' standalone='yes' ?>
<tagfile doxygen_version="1.9.8">
{template_all_nodes.rstrip()}
</tagfile>
"""
        with open(output_path, "w", encoding="utf8") as file:
            file.write(template_xml)



from typing import Callable, Dict, Optional, Tuple

from textx import metamodel_from_str

from strictdoc.core.query_engine.grammar import QUERY_GRAMMAR
from strictdoc.core.query_engine.query_object import (
    AndExpression,
    EqualExpression,
    InExpression,
    NodeContainsAnyFreeTextExpression,
    NodeContainsExpression,
    NodeFieldExpression,
    NodeHasChildRequirementsExpression,
    NodeHasParentRequirementsExpression,
    NodeIsRequirementExpression,
    NodeIsRootExpression,
    NodeIsSectionExpression,
    NodeIsSourceFileExpression,
    NodeIsSourceFileWithCompleteCoverageExpression,
    NodeIsSourceFileWithNoCoverageExpression,
    NodeIsSourceFileWithPartialCoverageExpression,
    NoneExpression,
    NotEqualExpression,
    NotExpression,
    NotInExpression,
    OrExpression,
    Query,
    StringExpression,
)

QUERY_MODELS = [
    AndExpression,
    EqualExpression,
    InExpression,
    NodeContainsExpression,
    NodeContainsAnyFreeTextExpression,
    NodeFieldExpression,
    NodeHasChildRequirementsExpression,
    NodeHasParentRequirementsExpression,
    NodeIsRequirementExpression,
    NodeIsRootExpression,
    NodeIsSectionExpression,
    NodeIsSourceFileExpression,
    NodeIsSourceFileWithCompleteCoverageExpression,
    NodeIsSourceFileWithPartialCoverageExpression,
    NodeIsSourceFileWithNoCoverageExpression,
    NoneExpression,
    NotEqualExpression,
    NotExpression,
    NotInExpression,
    OrExpression,
    Query,
    StringExpression,
]


class QueryParseContext:
    pass


class QueryParsingProcessor:
    def __init__(self, parse_context: QueryParseContext) -> None:
        self.parse_context: QueryParseContext = parse_context

    def get_default_processors(self) -> Dict[str, Callable[..., None]]:
        return {}


class QueryReader:
    def __init__(self, path_to_output_root: str = "NOT_RELEVANT") -> None:
        self.path_to_output_root = path_to_output_root

    @staticmethod
    def _read(
        input_string: str, file_path: Optional[str] = None
    ) -> Tuple[Query, QueryParseContext]:
        meta_model = metamodel_from_str(
            QUERY_GRAMMAR,
            classes=QUERY_MODELS,
            use_regexp_group=True,
        )

        parse_context = QueryParseContext()
        processor = QueryParsingProcessor(parse_context=parse_context)
        meta_model.register_obj_processors(processor.get_default_processors())

        query: Query = meta_model.model_from_str(
            input_string, file_name=file_path
        )

        return query, parse_context

    @staticmethod
    def read(input_string: str, file_path: Optional[str] = None) -> Query:
        document, _ = QueryReader._read(input_string, file_path)
        return document



import os
import platform
import sys


class SDocRuntimeEnvironment:
    def __init__(self, path_to_init: str) -> None:
        assert isinstance(path_to_init, str), path_to_init

        # FIXME: Delete me.
        self.path_to_init = path_to_init
        # https://stackoverflow.com/a/55556360/598057
        self.is_nuitka = "__compiled__" in globals()
        self.is_py_installer = (
            getattr(sys, "frozen", False) and not self.is_nuitka
        )

        # "frozen" attribute is set to True by PyInstaller, and it looks like
        # Nuitka does the same.
        self.is_binary_dist: bool = (
            getattr(sys, "frozen", False) or self.is_nuitka
        )

        if self.is_binary_dist:  # pragma: no cover
            # When it is a binary distribution, we don't have a Python package
            # file structure but only a binary with surrounding libraries and
            # data files (Jinja HTML files, CSS, JS, ...).
            path_to_main = os.path.abspath(sys.argv[0])
            path_to_main_dir = os.path.dirname(os.path.abspath(path_to_main))

            # Nuitka.
            if self.is_nuitka:
                self.path_to_strictdoc = path_to_main_dir
            # PyInstaller
            else:
                self.path_to_strictdoc = (
                    sys._MEIPASS  # type: ignore[attr-defined]
                )
        else:
            self.path_to_strictdoc = os.path.abspath(
                os.path.join(path_to_init, "..", "..")
            )
        assert os.path.isdir(self.path_to_strictdoc), self.path_to_strictdoc
        assert os.path.isabs(self.path_to_strictdoc), (
            "Path to strictdoc's package path must be an absolute path: "
            f"{self.path_to_strictdoc}"
        )

        self.is_debug_mode: bool = False

    def is_windows(self) -> bool:
        return platform.system() == "Windows"

    def is_github_ci_windows(self) -> bool:
        return self.is_windows() and os.environ.get("GITHUB_ACTIONS") == "true"

    def get_static_files_path(self) -> str:
        if self.is_binary_dist:  # pragma: no cover
            return os.path.join(self.path_to_strictdoc, "_static")
        return os.path.join(
            self.path_to_strictdoc, "strictdoc/export/html/_static"
        )

    def get_extra_static_files_path(self) -> str:
        if self.is_binary_dist:  # pragma: no cover
            return os.path.join(self.path_to_strictdoc, "_static_extra")
        return os.path.join(
            self.path_to_strictdoc, "strictdoc/export/html/_static_extra"
        )

    def get_path_to_rst_templates(self) -> str:
        if self.is_py_installer:  # pragma: no cover
            # If the application is run as a bundle, the PyInstaller bootloader
            # extends the sys module by a flag frozen=True and sets the app
            # path into variable _MEIPASS'.
            bundle_dir = (
                sys._MEIPASS  # type: ignore[attr-defined]
            )
            return os.path.join(bundle_dir, "templates/rst")
        if self.is_nuitka:  # pragma: no cover
            return os.path.join(self.path_to_strictdoc, "templates/rst")
        # Normal Python
        return os.path.join(
            self.path_to_strictdoc, "strictdoc", "export", "rst", "templates"
        )

    def get_path_to_html_templates(self) -> str:
        if self.is_py_installer:  # pragma: no cover
            # If the application is run as a bundle, the PyInstaller bootloader
            # extends the sys module by a flag frozen=True and sets the app
            # path into variable _MEIPASS'.
            bundle_dir = (
                sys._MEIPASS  # type: ignore[attr-defined]
            )
            return os.path.join(bundle_dir, "templates/html")
        if self.is_nuitka:  # pragma: no cover
            return os.path.join(self.path_to_strictdoc, "templates/html")
        # Normal Python.
        path_to_html_templates = os.path.join(
            self.path_to_strictdoc, "strictdoc", "export", "html", "templates"
        )
        assert os.path.isdir(path_to_html_templates), path_to_html_templates
        assert os.path.isabs(path_to_html_templates), path_to_html_templates
        return path_to_html_templates



"""
@relation(SDOC-SRS-30, scope=file)
"""

from typing import List


class DocumentTreeError(Exception):
    def __init__(self, problem_uid: str, cycled_uids: List[str]) -> None:
        super().__init__()
        self.problem_uid: str = problem_uid
        self.cycled_uids: List[str] = cycled_uids

    @staticmethod
    def cycle_error(
        problem_uid: str, cycled_uids: List[str]
    ) -> "DocumentTreeError":
        return DocumentTreeError(problem_uid, cycled_uids)

    def to_print_message(self) -> str:
        cycled_uids = ", ".join(self.cycled_uids)
        message = (
            "error: document tree: "
            "a cycle detected: requirements in the document tree must not "
            "reference each other.\n"
            f"Problematic UID: {self.problem_uid}.\nCycle: {cycled_uids}.\n"
        )
        return message

    def to_validation_message(self) -> str:
        return (
            "A cycle detected: requirements in the document tree must not "
            "reference each other. "
            f"Problematic UID: {self.problem_uid}, cycle: {self.cycled_uids}."
        )



from dataclasses import dataclass
from typing import List, Optional


@dataclass
class Metric:
    name: str
    value: str
    link: Optional[str] = None


@dataclass
class MetricSection:
    name: str
    metrics: List[Metric]



"""
@relation(SDOC-SRS-99, scope=file)
"""

from typing import List, Optional, Union

from strictdoc.backend.sdoc.document_reference import DocumentReference
from strictdoc.backend.sdoc.models.document_grammar import DocumentGrammar
from strictdoc.backend.sdoc.models.grammar_element import GrammarElement
from strictdoc.backend.sdoc.models.model import (
    SDocDocumentIF,
    SDocElementIF,
    SDocSectionIF,
)
from strictdoc.helpers.auto_described import auto_described
from strictdoc.helpers.cast import assert_cast
from strictdoc.helpers.mid import MID


@auto_described
class SectionContext:
    def __init__(self) -> None:
        self.title_number_string: Optional[str] = None


@auto_described
class SDocSection(SDocSectionIF):
    def __init__(
        self,
        parent: Union[SDocDocumentIF, SDocSectionIF],
        mid: Optional[str],
        uid: Optional[str],
        custom_level: Optional[str],
        title: str,
        requirement_prefix: Optional[str],
        section_contents: List[SDocElementIF],
    ):
        self.parent: Union[SDocDocumentIF, SDocSectionIF] = parent

        # TODO: Remove .uid, keep reserved_uid only.
        meaningful_uid: Optional[str] = None
        if uid is not None and len(uid) > 0:
            meaningful_uid = uid
        self.reserved_uid: Optional[str] = meaningful_uid

        self.title: str = title
        self.reserved_title = title
        self.requirement_prefix: Optional[str] = requirement_prefix

        self.section_contents = section_contents

        self.custom_level: Optional[str] = custom_level
        self.ng_resolved_custom_level: Optional[str] = custom_level

        self.ng_level: Optional[int] = None
        self.ng_has_requirements = False
        self.ng_document_reference: Optional[DocumentReference] = None
        self.ng_including_document_reference: Optional[DocumentReference] = None
        self.context = SectionContext()

        self.reserved_mid: MID = MID(mid) if mid is not None else MID.create()
        self.mid_permanent: bool = mid is not None

        # This is always true, unless the node is filtered out with --filter-requirements.
        self.ng_whitelisted = True
        self.ng_line_start: Optional[int] = None
        self.ng_line_end: Optional[int] = None
        self.ng_col_start: Optional[int] = None
        self.ng_col_end: Optional[int] = None
        self.ng_byte_start: Optional[int] = None
        self.ng_byte_end: Optional[int] = None

    @staticmethod
    def get_type_string() -> str:
        return "section"

    def get_display_node_type(self) -> str:
        return "Section"

    def get_node_type_string(self) -> Optional[str]:
        return None

    def get_debug_info(self) -> str:
        debug_components: List[str] = [f"TITLE = '{self.title}'"]
        document: Optional[SDocDocumentIF] = self.get_document()
        if document is not None:
            debug_components.append(f"document = {document.get_debug_info()}")
        return f"Section({', '.join(debug_components)})"

    @property
    def is_root_included_document(self) -> bool:
        return False

    def get_display_title(self, include_toc_number: bool = True) -> str:
        if include_toc_number and self.context.title_number_string is not None:
            return f"{self.context.title_number_string}. {self.title}"
        return self.title

    def get_document(self) -> Optional[SDocDocumentIF]:
        assert self.ng_document_reference is not None
        return self.ng_document_reference.get_document()

    def get_including_document(self) -> Optional[SDocDocumentIF]:
        assert self.ng_including_document_reference is not None
        return self.ng_including_document_reference.get_document()

    def get_parent_or_including_document(self) -> SDocDocumentIF:
        assert self.ng_including_document_reference is not None
        including_document_or_none = (
            self.ng_including_document_reference.get_document()
        )
        if including_document_or_none is not None:
            return including_document_or_none

        assert self.ng_document_reference is not None
        document: Optional[SDocDocumentIF] = (
            self.ng_document_reference.get_document()
        )
        assert document is not None, (
            "A valid requirement must always have a reference to the document."
        )
        return document

    def document_is_included(self) -> bool:
        assert self.ng_including_document_reference is not None
        return self.ng_including_document_reference.get_document() is not None

    def is_requirement(self) -> bool:
        return False

    def is_section(self) -> bool:
        return True

    def has_any_text_nodes(self) -> bool:
        # FIXME: Remove using __class__.
        return any(
            node_.__class__.__name__ == "SDocNode" and node_.node_type == "TEXT"  # type: ignore[union-attr]
            for node_ in self.section_contents
        )

    @property
    def is_root(self) -> bool:
        document: SDocDocumentIF = assert_cast(
            self.get_document(), SDocDocumentIF
        )
        return document.config.root is True

    def get_prefix(self) -> str:
        if self.requirement_prefix is not None:
            return self.requirement_prefix
        parent: Union[SDocSectionIF, SDocDocumentIF] = self.parent
        return parent.get_prefix()

    def get_prefix_for_new_node(self, node_type: str) -> Optional[str]:
        assert isinstance(node_type, str) and len(node_type), node_type

        document: SDocDocumentIF = assert_cast(
            self.get_document(), SDocDocumentIF
        )
        grammar: DocumentGrammar = assert_cast(
            document.grammar, DocumentGrammar
        )
        element: GrammarElement = grammar.elements_by_type[node_type]
        if (element_prefix := element.property_prefix) is not None:
            if element_prefix == "None":
                return None
            return element_prefix

        return self.get_prefix()

    def blacklist_if_needed(self) -> None:
        for node_ in self.section_contents:
            if node_.ng_whitelisted:
                return

        self.ng_whitelisted = False

        # If it turns out that all child nodes are blacklisted,
        # go up and blacklist the parent node if needed.
        if isinstance(self.parent, SDocSection) and self.parent.ng_whitelisted:
            self.parent.blacklist_if_needed()



import atexit
import os
import signal
import types
from typing import Optional


def register_code_coverage_hook() -> None:
    if not "COVERAGE_PROCESS_START" in os.environ:
        # This branch will never be checked by the coverage.
        return  # pragma: no cover

    import coverage  # noqa: PLC0415

    current_coverage = coverage.Coverage.current()

    assert current_coverage is not None

    def save_coverage() -> None:
        print(  # noqa: T201
            "strictdoc/server: exit hook: saving code coverage...",
            flush=True,
        )
        current_coverage.stop()
        # Code coverage is stopped at this point. Marking the next line to
        # be excluded from code coverage.
        current_coverage.save()  # pragma: no cover

    atexit.register(save_coverage)

    def handle_signal(
        signum: int,
        frame: Optional[types.FrameType],  # noqa: ARG001
    ) -> None:
        print(  # noqa: T201
            f"strictdoc: caught signal {signum}.", flush=True
        )
        save_coverage()
        # Code coverage is stopped at this point. Marking the next line to
        # be excluded from code coverage.
        signal.signal(signum, signal.SIG_DFL)  # pragma: no cover
        os.kill(os.getpid(), signum)  # pragma: no cover

    for sig in (signal.SIGINT, signal.SIGTERM):
        signal.signal(sig, handle_signal)



"""
@relation(SDOC-SRS-144, scope=file)
"""

import hashlib
import json
from dataclasses import dataclass
from pathlib import Path
from typing import Type

from strictdoc.backend.sdoc.document_reference import DocumentReference
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.document_grammar import DocumentGrammar
from strictdoc.backend.sdoc.models.node import SDocNode, SDocNodeField
from strictdoc.backend.sdoc.models.reference import (
    FileEntry,
    FileEntryFormat,
    FileReference,
)
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.core.file_tree import File
from strictdoc.core.project_config import ProjectConfig


@dataclass
class GCovStatsObject:
    total_number_of_covered_functions: int = 0
    total_number_of_non_covered_functions: int = 0

    def add_function(self, covered: bool) -> None:
        if covered:
            self.total_number_of_covered_functions += 1
        else:
            self.total_number_of_non_covered_functions += 1


class GCovJSONReader:
    @classmethod
    def read_from_file(
        cls: Type["GCovJSONReader"],
        doc_file: File,
        project_config: ProjectConfig,
    ) -> SDocDocument:
        with open(doc_file.full_path, encoding="UTF-8") as file:
            content = file.read()
        return cls.read_from_string(content, doc_file, project_config)

    @classmethod
    def read_from_string(
        cls: Type["GCovJSONReader"],
        content: str,
        doc_file: File,  # noqa: ARG003
        project_config: ProjectConfig,  # noqa: ARG003
    ) -> SDocDocument:
        if len(content) == 0:
            raise RuntimeError("Document is empty")
        try:
            json_content = json.loads(content)
        except Exception as exception:  # pylint: disable=broad-except
            raise RuntimeError(str(exception)) from None

        document = SDocDocument(
            mid=None,
            title="Code coverage report",
            config=None,
            view=None,
            grammar=None,
            section_contents=[],
        )
        document.ng_including_document_reference = DocumentReference()
        grammar = DocumentGrammar.create_for_test_report(document)
        document.grammar = grammar
        document.config.requirement_style = "Table"

        stats = GCovStatsObject()

        """
        Parse individual <testcase> elements.
        """

        json_files = json_content["files"]
        for json_file_ in json_files:
            json_file_name = json_file_["file"]

            file_section = SDocSection(
                parent=document,
                mid=None,
                uid=None,
                custom_level=None,
                title=json_file_name,
                requirement_prefix=None,
                section_contents=[],
            )
            file_section.ng_including_document_reference = DocumentReference()
            file_section.ng_document_reference = DocumentReference()
            file_section.ng_document_reference.set_document(document)
            document.section_contents.append(file_section)

            covered_functions_section = SDocSection(
                parent=document,
                mid=None,
                uid=None,
                custom_level=None,
                title="Covered functions",
                requirement_prefix=None,
                section_contents=[],
            )
            covered_functions_section.ng_including_document_reference = (
                DocumentReference()
            )
            covered_functions_section.ng_document_reference = (
                DocumentReference()
            )
            covered_functions_section.ng_document_reference.set_document(
                document
            )
            file_section.section_contents.append(covered_functions_section)

            non_covered_functions_section = SDocSection(
                parent=document,
                mid=None,
                uid=None,
                custom_level=None,
                title="Non-covered functions",
                requirement_prefix=None,
                section_contents=[],
            )
            non_covered_functions_section.ng_including_document_reference = (
                DocumentReference()
            )
            non_covered_functions_section.ng_document_reference = (
                DocumentReference()
            )
            non_covered_functions_section.ng_document_reference.set_document(
                document
            )
            file_section.section_contents.append(non_covered_functions_section)

            for json_function_ in json_file_["functions"]:
                json_function_name = json_function_["demangled_name"]
                is_function_covered = json_function_["execution_count"] > 0
                stats.add_function(is_function_covered)

                testcase_node = SDocNode(
                    parent=document,
                    node_type="TEST_RESULT",
                    fields=[],
                    relations=[],
                )
                testcase_node.ng_document_reference = DocumentReference()
                testcase_node.ng_document_reference.set_document(document)
                testcase_node.ng_including_document_reference = (
                    DocumentReference()
                )
                testcase_node.set_field_value(
                    field_name="UID",
                    form_field_index=0,
                    value=SDocNodeField(
                        parent=testcase_node,
                        field_name="UID",
                        parts=[
                            "GCOV:" + json_file_name + ":" + json_function_name
                        ],
                        multiline__=None,
                    ),
                )
                # FIXME: Remove?
                testcase_node.set_field_value(
                    field_name="STATUS",
                    form_field_index=0,
                    value=SDocNodeField(
                        parent=testcase_node,
                        field_name="STATUS",
                        parts=[
                            "Covered" if is_function_covered else "Non-covered"
                        ],
                        multiline__=None,
                    ),
                )

                path = "src/main.c"
                gcov_path_hash = hashlib.md5(path.encode("utf-8")).hexdigest()
                # FIXME: Resolve the relative path from a project config.
                link = (
                    "../../../../"  # noqa: ISC003
                    + "coverage."
                    + Path(json_file_name).name
                    + ". "
                    + gcov_path_hash
                    + ".html"
                )
                testcase_node.set_field_value(
                    field_name="STATEMENT",
                    form_field_index=0,
                    value=SDocNodeField(
                        parent=testcase_node,
                        field_name="STATEMENT",
                        parts=[
                            f"""
                            `LINK <{link}>`_
                            """
                        ],
                        multiline__=None,
                    ),
                )
                testcase_node.set_field_value(
                    field_name="TITLE",
                    form_field_index=0,
                    value=SDocNodeField(
                        parent=testcase_node,
                        field_name="TITLE",
                        parts=[json_function_name],
                        multiline__=None,
                    ),
                )
                testcase_node.relations.append(
                    FileReference(
                        parent=testcase_node,
                        g_file_entry=FileEntry(
                            parent=None,
                            g_file_format=FileEntryFormat.SOURCECODE,
                            g_file_path=json_file_name,
                            g_line_range=None,
                            function=json_function_name,
                            clazz=None,
                        ),
                    )
                )
                if is_function_covered:
                    covered_functions_section.section_contents.append(
                        testcase_node
                    )
                else:
                    non_covered_functions_section.section_contents.append(
                        testcase_node
                    )

            summary_table = f"""\
.. list-table:: Coverage summary
    :widths: 25 10
    :header-rows: 0

    * - **Number of files:**
      - {len(json_content["files"])}
    * - **Total covered functions:**
      - {stats.total_number_of_covered_functions}
    * - **Total non-covered functions:**
      - {stats.total_number_of_non_covered_functions}
        """

            testcase_node = SDocNode(
                parent=document,
                node_type="TEXT",
                fields=[],
                relations=[],
            )
            testcase_node.ng_document_reference = DocumentReference()
            testcase_node.ng_document_reference.set_document(document)
            testcase_node.ng_including_document_reference = DocumentReference()
            testcase_node.set_field_value(
                field_name="STATEMENT",
                form_field_index=0,
                value=SDocNodeField(
                    parent=testcase_node,
                    field_name="STATEMENT",
                    parts=[summary_table],
                    multiline__="True",
                ),
            )
            document.section_contents.insert(0, testcase_node)

        return document



from typing import Iterator, List, Optional, Tuple, Union

from markupsafe import Markup

from strictdoc import __version__
from strictdoc.backend.sdoc.models.anchor import Anchor
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.node import SDocNode
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.core.document_tree_iterator import DocumentTreeIterator
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.export.html.html_templates import JinjaEnvironment
from strictdoc.export.html.renderers.link_renderer import LinkRenderer
from strictdoc.export.html.renderers.markup_renderer import MarkupRenderer


class TraceabilityMatrixViewObject:
    def __init__(
        self,
        *,
        traceability_index: TraceabilityIndex,
        project_config: ProjectConfig,
        link_renderer: LinkRenderer,
        markup_renderer: MarkupRenderer,
        known_relations_list: List[Tuple[str, Optional[str]]],
    ) -> None:
        self.traceability_index: TraceabilityIndex = traceability_index
        self.project_config: ProjectConfig = project_config
        self.link_renderer: LinkRenderer = link_renderer
        self.markup_renderer: MarkupRenderer = markup_renderer
        self.known_relations_list: List[Tuple[str, Optional[str]]] = (
            known_relations_list
        )
        self.standalone: bool = False
        self.document_tree_iterator: DocumentTreeIterator = (
            DocumentTreeIterator(traceability_index.document_tree)
        )
        self.is_running_on_server: bool = project_config.is_running_on_server
        self.strictdoc_version = __version__

    def iterate_documents(self) -> Iterator[SDocDocument]:
        yield from filter(
            lambda document_: not document_.document_is_included(),
            self.traceability_index.document_tree.document_list,
        )

    def render_screen(self, jinja_environment: JinjaEnvironment) -> Markup:
        return jinja_environment.render_template_as_markup(
            "screens/traceability_matrix/index.jinja", view_object=self
        )

    def render_static_url(self, url: str) -> str:
        return self.link_renderer.render_static_url(url)

    def render_url(self, url: str) -> str:
        # FIXME: Return Markup(...).
        return self.link_renderer.render_url(url)

    def render_local_anchor(
        self, node: Union[Anchor, SDocNode, SDocSection, SDocDocument]
    ) -> str:
        return self.link_renderer.render_local_anchor(node)

    def get_document_level(self) -> int:
        return 0



"""
@relation(SDOC-SRS-33, SDOC-SRS-142, scope=file)
"""

from typing import Optional, Union

from strictdoc.backend.sdoc.pickle_cache import PickleCache
from strictdoc.backend.sdoc_source_code.models.source_file_info import (
    SourceFileTraceabilityInfo,
)
from strictdoc.backend.sdoc_source_code.reader import (
    SourceFileTraceabilityReader,
)
from strictdoc.backend.sdoc_source_code.reader_c import (
    SourceFileTraceabilityReader_C,
)
from strictdoc.backend.sdoc_source_code.reader_python import (
    SourceFileTraceabilityReader_Python,
)
from strictdoc.backend.sdoc_source_code.reader_robot import (
    SourceFileTraceabilityReader_Robot,
)
from strictdoc.core.project_config import ProjectConfig


class SourceFileTraceabilityCachingReader:
    @staticmethod
    def read_from_file(
        path_to_file: str, project_config: ProjectConfig
    ) -> Optional[SourceFileTraceabilityInfo]:
        unpickled_content = PickleCache.read_from_cache(
            path_to_file, project_config, "source_file"
        )
        if unpickled_content is not None:
            assert isinstance(unpickled_content, SourceFileTraceabilityInfo), (
                path_to_file,
                unpickled_content,
            )
            return unpickled_content

        reader = SourceFileTraceabilityCachingReader._get_reader(
            path_to_file, project_config
        )
        try:
            traceability_info = reader.read_from_file(path_to_file)
            PickleCache.save_to_cache(
                traceability_info,
                path_to_file,
                project_config,
                "source_file",
            )
        except UnicodeDecodeError:
            print(  # noqa: T201
                f"warning: Skip tracing binary file {path_to_file}."
            )
            return None

        return traceability_info

    @staticmethod
    def _get_reader(
        path_to_file: str, project_config: ProjectConfig
    ) -> Union[
        SourceFileTraceabilityReader,
        SourceFileTraceabilityReader_Python,
        SourceFileTraceabilityReader_C,
        SourceFileTraceabilityReader_Robot,
    ]:
        if project_config.is_activated_source_file_language_parsers():
            if path_to_file.endswith(".py"):
                return SourceFileTraceabilityReader_Python()
            if (
                path_to_file.endswith(".c")
                or path_to_file.endswith(".cc")
                or path_to_file.endswith(".h")
                or path_to_file.endswith(".hh")
                or path_to_file.endswith(".hpp")
                or path_to_file.endswith(".cpp")
            ):
                parse_nodes = project_config.shall_parse_nodes(path_to_file)
                return SourceFileTraceabilityReader_C(parse_nodes=parse_nodes)
            if path_to_file.endswith(".robot"):
                return SourceFileTraceabilityReader_Robot()
        return SourceFileTraceabilityReader()



from markupsafe import Markup


class HTMLFragmentWriter:
    @staticmethod
    def write(text_fragment: str) -> Markup:
        return Markup(text_fragment)

    @staticmethod
    def write_anchor_link(title: str, href: str) -> str:
        return f'<a href="{href}">🔗&nbsp;{title}</a>`'



"""This module provides a helper removeprefix. (compatibility with python 3.8)"""

import sys


# TODO: Once requires_python is set to 3.9, replace with str.removeprefix() and remove this helper.
def removeprefix(s: str, prefix: str) -> str:
    if sys.version_info >= (3, 9):
        return s.removeprefix(prefix)
    if s.startswith(prefix):
        return s[len(prefix) :]
    return s



"""
Run StrictDoc child tasks as separate processes.

@relation(SDOC-SRS-1, scope=file)
"""

import multiprocessing
from abc import ABC, abstractmethod
from concurrent.futures import ProcessPoolExecutor, as_completed
from typing import Any, Callable, Iterable, List, Optional, Tuple

from strictdoc import environment
from strictdoc.helpers.exception import (
    ExceptionInfo,
    StrictDocChildProcessException,
)

MultiprocessingLambdaType = Callable[[Any], Any]


def processing_func_wrapper(
    func: MultiprocessingLambdaType, input_arg: Any
) -> Tuple[Optional[Any], Optional[StrictDocChildProcessException]]:
    try:
        result = func(input_arg)
        return result, None
    except Exception as exception_:
        return None, StrictDocChildProcessException(ExceptionInfo(exception_))


class Parallelizer(ABC):
    @staticmethod
    def create(parallelize: bool) -> "Parallelizer":
        if parallelize:
            return MultiprocessingParallelizer()
        return NullParallelizer()

    @abstractmethod
    def run_parallel(
        self,
        contents: List[Any],
        processing_func: MultiprocessingLambdaType,
    ) -> Iterable[Any]:
        raise NotImplementedError

    @abstractmethod
    def shutdown(self) -> None:
        raise NotImplementedError


class MultiprocessingParallelizer(Parallelizer):
    def __init__(self) -> None:
        process_number: int = multiprocessing.cpu_count()

        if environment.is_github_ci_windows():  # pragma: no cover
            fixed_process_number = 2
            print(  # noqa: T201
                f"MultiprocessingParallelizer: "
                f"Running on GitHub CI Windows with only "
                f"{fixed_process_number} parallel processes instead of "
                f"{process_number}."
            )
            process_number = 2

        self.executor = ProcessPoolExecutor(max_workers=process_number)

    def run_parallel(
        self,
        contents: List[Any],
        processing_func: MultiprocessingLambdaType,
    ) -> Iterable[Any]:
        try:
            future_to_index = {
                self.executor.submit(
                    processing_func_wrapper, processing_func, item
                ): idx
                for idx, item in enumerate(contents)
            }
            results = [None] * len(contents)

            for future in as_completed(future_to_index):
                idx = future_to_index[future]
                result = future.result()
                if result[1] is not None:
                    raise result[1]
                results[idx] = result[0]

            return results
        except Exception as e:
            raise e

    def shutdown(self) -> None:
        self.executor.shutdown(wait=True)


class NullParallelizer(Parallelizer):
    def run_parallel(
        self,
        contents: List[Any],
        processing_func: MultiprocessingLambdaType,
    ) -> Iterable[Any]:
        results = []
        for content in contents:
            results.append(processing_func(content))
        return results

    def shutdown(self) -> None:
        pass



"""
@relation(SDOC-SRS-152, scope=file)
"""

from enum import Enum
from typing import List

import openpyxl
import xlrd

from strictdoc.helpers.auto_described import auto_described
from strictdoc.helpers.cast import assert_cast


class ExcelLibType(Enum):
    XLRD = 1
    OPENPYXL = 2


@auto_described
class ExcelSheetProxy:
    """
    An adapter class to open xls(xlrd) or xlsx(openpyxl) with the same interface.
    """

    def __init__(self, file: str):
        self.file = file
        if file.endswith(".xlsx"):
            self.workbook = openpyxl.load_workbook(file)
            self.sheet = self.workbook.active
            if self.sheet is None:  # pragma: no cover
                raise RuntimeError(
                    "Couldn't open the first sheet of the workbook"
                )
            self.lib = ExcelLibType.OPENPYXL
        elif file.endswith(".xls"):
            self.workbook = xlrd.open_workbook(file, on_demand=True)
            self.sheet = self.workbook.sheet_by_index(0)
            self.lib = ExcelLibType.XLRD
        else:  # pragma: no cover
            raise AssertionError("Unsupported file format")

    @property
    def name(self) -> str:
        """
        Returns the name of the first sheet.
        """
        if self.lib == ExcelLibType.OPENPYXL:
            return str(self.sheet.title)
        elif self.lib == ExcelLibType.XLRD:
            return str(self.sheet.name)
        raise AssertionError

    @property
    def ncols(self) -> int:
        """
        The number of columns.
        """
        if self.lib == ExcelLibType.OPENPYXL:
            return assert_cast(self.sheet.max_column, int)
        elif self.lib == ExcelLibType.XLRD:
            return assert_cast(self.sheet.ncols, int)
        raise AssertionError

    @property
    def nrows(self) -> int:
        """
        The number of rows.
        """
        if self.lib == ExcelLibType.OPENPYXL:
            return assert_cast(self.sheet.max_row, int)
        elif self.lib == ExcelLibType.XLRD:
            return assert_cast(self.sheet.nrows, int)
        raise AssertionError

    def get_cell_value(self, row: int, col: int) -> str:
        """
        Returns the value at row/col as string.
        """
        if self.lib == ExcelLibType.OPENPYXL:
            return self.sheet.cell(row=row + 1, column=col + 1).value or ""
        elif self.lib == ExcelLibType.XLRD:
            return assert_cast(self.sheet.cell_value(row, col), str)
        raise AssertionError

    def row_values(self, row: int) -> List[str]:
        """
        Returns a full row as list.
        """
        if self.lib == ExcelLibType.OPENPYXL:
            return [(cell.value or "") for cell in self.sheet[row + 1]]
        elif self.lib == ExcelLibType.XLRD:
            return assert_cast(self.sheet.row_values(row), list)
        raise AssertionError



from markupsafe import Markup


def render_turbo_stream(content: str, action: str, target: str) -> Markup:
    assert action in ("append", "replace", "update")

    turbo_stream = f"""
<turbo-stream action="{action}" target="{target}">
  <template>
    {content}
  </template>
</turbo-stream>
"""
    return Markup(turbo_stream)



from copy import copy
from typing import Any, Optional, Tuple

from textx import TextXSemanticError, TextXSyntaxError, metamodel_from_str

from strictdoc.backend.sdoc.grammar.grammar_builder import SDocGrammarBuilder
from strictdoc.backend.sdoc.models.constants import DOCUMENT_MODELS
from strictdoc.backend.sdoc.models.document import (
    SDocDocument,
)
from strictdoc.backend.sdoc.models.document_grammar import DocumentGrammar
from strictdoc.backend.sdoc.models.node import (
    SDocCompositeNode,
    SDocNode,
    SDocNodeField,
)
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.backend.sdoc.pickle_cache import PickleCache
from strictdoc.backend.sdoc.processor import ParseContext, SDocParsingProcessor
from strictdoc.core.project_config import ProjectConfig
from strictdoc.helpers.cast import assert_cast
from strictdoc.helpers.exception import StrictDocException
from strictdoc.helpers.textx import drop_textx_meta


class SDReader:
    meta_model = metamodel_from_str(
        SDocGrammarBuilder.create_grammar(),
        classes=DOCUMENT_MODELS,
        use_regexp_group=True,
    )

    @staticmethod
    def _read(
        input_string: str,
        file_path: Optional[str] = None,
        migrate_sections: bool = False,
    ) -> Tuple[SDocDocument, ParseContext]:
        parse_context = ParseContext(
            path_to_sdoc_file=file_path, migrate_sections=migrate_sections
        )
        processor = SDocParsingProcessor(parse_context=parse_context)
        SDReader.meta_model.register_obj_processors(
            processor.get_default_processors()
        )

        try:
            document: SDocDocument = SDReader.meta_model.model_from_str(
                input_string, file_name=file_path
            )
        except (TextXSyntaxError, TextXSemanticError) as syntax_error_:
            raise StrictDocException(
                f"Could not parse file: "
                f"{file_path}. "
                f"Error: {syntax_error_.__class__.__name__}: {syntax_error_}"
            ) from syntax_error_

        parse_context.document_reference.set_document(document)
        document.ng_has_requirements = parse_context.document_has_requirements

        return document, parse_context

    @staticmethod
    def read(
        input_string: str,
        file_path: Optional[str] = None,
        migrate_sections: bool = False,
    ) -> SDocDocument:
        document, _ = SDReader.read_with_parse_context(
            input_string, file_path, migrate_sections=migrate_sections
        )
        return document

    @staticmethod
    def read_with_parse_context(
        input_string: str,
        file_path: Optional[str] = None,
        migrate_sections: bool = False,
    ) -> Tuple[SDocDocument, ParseContext]:
        document, parse_context = SDReader._read(input_string, file_path)

        # FIXME: When the [SECTION] is gone, remove this.
        if migrate_sections:
            SDReader.migrate_sections(document)
            SDReader.migration_sections_grammar(document)
        SDReader.fixup_composite_requirements(document)

        return document, parse_context

    def read_from_file(
        self, file_path: str, project_config: ProjectConfig
    ) -> SDocDocument:
        """
        Parse a provided .sdoc file and convert it into a Document object.
        """

        unpickled_content = PickleCache.read_from_cache(
            file_path, project_config, "sdoc"
        )
        if unpickled_content:
            return assert_cast(unpickled_content, SDocDocument)

        with open(file_path, encoding="utf8") as file:
            sdoc_content = file.read()

        sdoc, parse_context = self.read_with_parse_context(
            sdoc_content,
            file_path=file_path,
            migrate_sections=project_config.is_new_section_behavior(),
        )

        sdoc.fragments_from_files = parse_context.fragments_from_files

        # HACK:
        # ProcessPoolExecutor doesn't work because of non-picklable parts
        # of textx. The offending fields are stripped down because they
        # are not used anyway.
        drop_textx_meta(sdoc)

        # @relation(SDOC-SRS-155, scope=range_start)
        sdoc.build_search_index()
        # @relation(SDOC-SRS-155, scope=range_end)

        PickleCache.save_to_cache(sdoc, file_path, project_config, "sdoc")

        return sdoc

    @staticmethod
    def convert(section: SDocSection) -> SDocNode:
        fields = []

        if section.mid_permanent:
            fields.append(
                SDocNodeField.create_from_string(
                    None,
                    field_name="MID",
                    field_value=section.reserved_mid,
                    multiline=False,
                )
            )
        if section.reserved_uid is not None:
            fields.append(
                SDocNodeField.create_from_string(
                    None,
                    field_name="UID",
                    field_value=section.reserved_uid,
                    multiline=False,
                )
            )
        if section.ng_resolved_custom_level is not None:
            fields.append(
                SDocNodeField.create_from_string(
                    None,
                    field_name="LEVEL",
                    field_value=section.ng_resolved_custom_level,
                    multiline=False,
                )
            )
        if (
            section.requirement_prefix is not None
            and len(section.requirement_prefix) > 0
        ):
            fields.append(
                SDocNodeField.create_from_string(
                    None,
                    field_name="PREFIX",
                    field_value=section.requirement_prefix,
                    multiline=False,
                )
            )
        fields.append(
            SDocNodeField.create_from_string(
                None,
                field_name="TITLE",
                field_value=section.reserved_title,
                multiline=False,
            )
        )
        node: SDocNode = SDocNode(
            parent=section.parent,
            node_type="SECTION",
            fields=fields,
            relations=[],
            is_composite=True,
            section_contents=section.section_contents,
            node_type_close="SECTION",
        )
        for field_ in fields:
            field_.parent = node

        node.ng_including_document_reference = (
            section.ng_including_document_reference
        )
        node.ng_document_reference = section.ng_document_reference
        node.ng_level = section.ng_level
        node.ng_resolved_custom_level = section.ng_resolved_custom_level
        node.ng_line_start = section.ng_line_start
        node.ng_line_end = section.ng_line_end
        node.ng_col_start = section.ng_col_start
        node.ng_col_end = section.ng_col_end
        node.ng_byte_start = section.ng_byte_start
        node.ng_byte_end = section.ng_byte_end
        return node

    @staticmethod
    def fixup_composite_requirements(sdoc: SDocDocument) -> None:
        for _, node_ in enumerate(copy(sdoc.section_contents)):
            if isinstance(node_, SDocCompositeNode):
                SDReader.migration_sections_grammar(sdoc)
                break

    @staticmethod
    def migrate_sections(sdoc: Any) -> None:
        for node_idx_, node_ in enumerate(copy(sdoc.section_contents)):
            if isinstance(node_, SDocSection):
                SDReader.migrate_sections(node_)

                new_node = SDReader.convert(node_)
                sdoc.section_contents[node_idx_] = new_node

    @staticmethod
    def migration_sections_grammar(sdoc: SDocDocument) -> None:
        grammar: DocumentGrammar = assert_cast(sdoc.grammar, DocumentGrammar)
        section_element_exists = any(
            element_.tag == "SECTION" for element_ in grammar.elements
        )
        if not section_element_exists:
            grammar.update_with_elements(
                [
                    DocumentGrammar.create_default_section_element(
                        grammar, enable_mid=sdoc.config.enable_mid or False
                    )
                ]
                + grammar.elements
            )



"""
@relation(SDOC-SRS-28, SDOC-SRS-33, scope=file)
"""

from copy import copy
from typing import (
    TYPE_CHECKING,
    Dict,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)

from strictdoc.backend.sdoc.document_reference import DocumentReference
from strictdoc.backend.sdoc.error_handling import StrictDocSemanticError
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.model import SDocDocumentIF, SDocNodeIF
from strictdoc.backend.sdoc.models.node import SDocNode
from strictdoc.backend.sdoc.models.reference import FileEntry, FileReference
from strictdoc.backend.sdoc_source_code.models.function import Function
from strictdoc.backend.sdoc_source_code.models.function_range_marker import (
    ForwardFunctionRangeMarker,
    FunctionRangeMarker,
    RangeMarkerType,
)
from strictdoc.backend.sdoc_source_code.models.line_marker import LineMarker
from strictdoc.backend.sdoc_source_code.models.range_marker import (
    ForwardFileMarker,
    ForwardRangeMarker,
    RangeMarker,
)
from strictdoc.backend.sdoc_source_code.models.requirement_marker import Req
from strictdoc.backend.sdoc_source_code.models.source_file_info import (
    RelationMarkerType,
    SourceFileTraceabilityInfo,
)
from strictdoc.core.constants import GraphLinkType
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.source_tree import SourceFile
from strictdoc.helpers.cast import assert_cast
from strictdoc.helpers.exception import StrictDocException
from strictdoc.helpers.google_test import convert_function_name_to_gtest_macro
from strictdoc.helpers.ordered_set import OrderedSet
from strictdoc.helpers.removeprefix import removeprefix

if TYPE_CHECKING:
    from strictdoc.core.traceability_index import (
        TraceabilityIndex,
    )


class FileTraceabilityIndex:
    def __init__(self) -> None:
        # "file.py" -> List[SDocNode]
        self.map_paths_to_reqs: Dict[str, OrderedSet[SDocNode]] = {}

        # "REQ-001" -> {"file.py", ...}
        self.map_reqs_uids_to_paths: Dict[str, OrderedSet[str]] = {}

        # "file.py" -> SourceFileTraceabilityInfo.
        self.map_paths_to_source_file_traceability_info: Dict[
            str, SourceFileTraceabilityInfo
        ] = {}

        # "file.py" -> { { "foo" -> [("REQ-1", "Impl"), ("REQ-2", "Test")] }, ... }
        self.map_file_function_names_to_reqs_uids: Dict[
            str, Dict[str, List[Tuple[str, Optional[str]]]]
        ] = {}
        self.map_file_class_names_to_reqs_uids: Dict[
            str, Dict[str, List[Tuple[str, Optional[str]]]]
        ] = {}

        # This is only public non-static functions from languages like C.
        self.map_all_function_names_to_definition_functions: Dict[
            str, List[Function]
        ] = {}

        # "file.py" -> [SDocNode]  # noqa: ERA001
        self.source_file_reqs_cache: Dict[str, Optional[List[SDocNode]]] = {}

        self.requirements_with_forward_links: OrderedSet[SDocNode] = (
            OrderedSet()
        )
        self.trace_infos: List[SourceFileTraceabilityInfo] = []

    def has_source_file_reqs(self, source_file_rel_path: str) -> bool:
        path_reqs = self.map_paths_to_reqs.get(source_file_rel_path)
        if path_reqs is not None and len(path_reqs) > 0:
            return True
        file_trace_info = self.map_paths_to_source_file_traceability_info[
            source_file_rel_path
        ]
        return len(file_trace_info.markers) > 0

    def get_requirement_file_links(
        self, requirement: SDocNode
    ) -> List[Tuple[str, List[RelationMarkerType]]]:
        if requirement.reserved_uid not in self.map_reqs_uids_to_paths:
            return []

        matching_links_with_markers: List[
            Tuple[str, List[RelationMarkerType]]
        ] = []
        requirement_source_paths: OrderedSet[str] = self.map_reqs_uids_to_paths[
            requirement.reserved_uid
        ]

        # Now that one requirement can have multiple File-relations to the same file.
        # This can be multiple FUNCTION: or RANGE: forward-relations.
        # To avoid duplication of results, visit each unique file link path only once.
        visited_file_links: Set[str] = set()
        for requirement_source_path_ in requirement_source_paths:
            if requirement_source_path_ in visited_file_links:
                continue
            visited_file_links.add(requirement_source_path_)

            source_file_traceability_info: Optional[
                SourceFileTraceabilityInfo
            ] = self.map_paths_to_source_file_traceability_info.get(
                requirement_source_path_
            )
            assert source_file_traceability_info is not None, (
                f"Requirement {requirement.reserved_uid} references a file"
                f" that does not exist: {requirement_source_path_}."
            )
            markers = source_file_traceability_info.ng_map_reqs_to_markers.get(
                requirement.reserved_uid
            )
            if markers is None or len(markers) == 0:
                matching_links_with_markers.append(
                    (requirement_source_path_, [])
                )
                continue
            matching_links_with_markers.append(
                (requirement_source_path_, markers)
            )

        return matching_links_with_markers

    def indexed_source_files(self) -> Iterator[SourceFile]:
        for _, sfti in self.map_paths_to_source_file_traceability_info.items():
            if sfti.source_file is not None:
                yield sfti.source_file

    def get_source_file_reqs(
        self, source_file_rel_path: str
    ) -> Optional[List[SDocNode]]:
        assert (
            source_file_rel_path
            in self.map_paths_to_source_file_traceability_info
        )
        if source_file_rel_path in self.source_file_reqs_cache:
            return self.source_file_reqs_cache[source_file_rel_path]

        source_file_traceability_info: SourceFileTraceabilityInfo = (
            self.map_paths_to_source_file_traceability_info[
                source_file_rel_path
            ]
        )

        if source_file_rel_path not in self.map_paths_to_reqs:
            self.source_file_reqs_cache[source_file_rel_path] = None
            return None

        requirements = self.map_paths_to_reqs[source_file_rel_path]
        assert len(requirements) > 0
        range_requirements = []

        for requirement in requirements:
            if (
                requirement.reserved_uid
                in source_file_traceability_info.ng_map_reqs_to_markers
            ):
                range_requirements.append(requirement)

        self.source_file_reqs_cache[source_file_rel_path] = range_requirements
        return range_requirements

    def get_coverage_info(
        self, source_file_rel_path: str
    ) -> SourceFileTraceabilityInfo:
        assert (
            source_file_rel_path
            in self.map_paths_to_source_file_traceability_info
        )
        source_file_tr_info: SourceFileTraceabilityInfo = (
            self.map_paths_to_source_file_traceability_info[
                source_file_rel_path
            ]
        )
        return source_file_tr_info

    def get_coverage_info_weak(
        self, source_file_rel_path: str
    ) -> Optional[SourceFileTraceabilityInfo]:
        source_file_tr_info: Optional[SourceFileTraceabilityInfo] = (
            self.map_paths_to_source_file_traceability_info.get(
                source_file_rel_path
            )
        )
        return source_file_tr_info

    def validate_and_resolve(
        self,
        traceability_index: "TraceabilityIndex",
        project_config: ProjectConfig,
    ) -> None:
        """
        Resolve all source code traceability after the index is fully built.
        """

        #
        # STEP: Collect minimal information that will help to resolve the
        #       forward-declared paths/function names at the step 2.
        #
        for trace_info_ in self.trace_infos:
            source_file: SourceFile = assert_cast(
                trace_info_.source_file, SourceFile
            )

            self.map_paths_to_source_file_traceability_info[
                source_file.in_doctree_source_file_rel_path_posix
            ] = trace_info_

            for function_ in trace_info_.functions:
                if function_.is_definition() and function_.is_public():
                    self.map_all_function_names_to_definition_functions.setdefault(
                        function_.name, []
                    ).append(function_)

        #
        # STEP: Resolve requirements that have forward links.
        #       Some requirements can come from the SDoc documents generated
        #       on the fly from JUnit XML documents.
        #
        for forward_requirement_ in self.requirements_with_forward_links:
            assert forward_requirement_.reserved_uid is not None

            for relation_ in forward_requirement_.relations:
                if not isinstance(relation_, FileReference):
                    continue

                file_reference: FileReference = assert_cast(
                    relation_, FileReference
                )
                file_posix_path = file_reference.get_posix_path()

                if file_posix_path == "#FORWARD#":
                    test_function = (
                        forward_requirement_.get_meta_field_value_by_title(
                            "TEST_FUNCTION"
                        )
                    )
                    assert test_function is not None

                    functions: List[Function]
                    if test_function.startswith("#GTEST#"):
                        test_function = removeprefix(test_function, "#GTEST#")
                        possible_gtest_functions = (
                            convert_function_name_to_gtest_macro(test_function)
                        )
                        for (
                            possible_gtest_function_
                        ) in possible_gtest_functions:
                            if (
                                possible_gtest_function_
                                in self.map_all_function_names_to_definition_functions
                            ):
                                test_function = possible_gtest_function_
                                break
                        else:
                            raise RuntimeError(
                                "Could not find a matching Google Test function: "
                                f"{possible_gtest_functions}"
                            )  # pragma: no cover
                        forward_requirement_.set_field_value(
                            field_name="TEST_FUNCTION",
                            form_field_index=0,
                            value=test_function,
                        )
                    functions = (
                        self.map_all_function_names_to_definition_functions[
                            test_function
                        ]
                    )
                    assert len(functions) == 1

                    function: Function = functions[0]
                    resolved_path_to_function_file = function.parent.source_file.in_doctree_source_file_rel_path_posix
                    file_posix_path = resolved_path_to_function_file

                    file_reference.g_file_entry = FileEntry(
                        relation_,
                        g_file_format=relation_.g_file_entry.g_file_format,
                        g_file_path=resolved_path_to_function_file,
                        g_line_range=None,
                        function=test_function,
                        clazz=None,
                    )

                    forward_requirement_.set_field_value(
                        field_name="TEST_PATH",
                        form_field_index=0,
                        value=resolved_path_to_function_file,
                    )

                    #
                    # This transitively connects requirements and test results
                    # through the test source files.
                    #
                    for function_marker_ in function.markers:
                        for req_ in function_marker_.reqs:
                            node = traceability_index.get_node_by_uid_weak2(
                                req_
                            )
                            traceability_index.graph_database.create_link(
                                link_type=GraphLinkType.NODE_TO_PARENT_NODES,
                                lhs_node=forward_requirement_,
                                rhs_node=node,
                                edge="Satisfies",
                            )
                            traceability_index.graph_database.create_link(
                                link_type=GraphLinkType.NODE_TO_CHILD_NODES,
                                lhs_node=node,
                                rhs_node=forward_requirement_,
                                edge="IsSatisfiedBy",
                            )
                #
                # Validate that all requirements reference existing files.
                #
                source_file_traceability_info: Optional[
                    SourceFileTraceabilityInfo
                ] = self.map_paths_to_source_file_traceability_info.get(
                    file_posix_path
                )
                if source_file_traceability_info is None:
                    raise StrictDocException(
                        f"Requirement {forward_requirement_.reserved_uid} "
                        "references a file that does not exist: "
                        f"{file_posix_path}."
                    )

                #
                # Now that the test reports related fixups are done, the
                # following code registers the requirements with forward links.
                #
                self.map_paths_to_reqs.setdefault(
                    file_posix_path, OrderedSet()
                ).add(forward_requirement_)

                assert forward_requirement_.reserved_uid is not None
                self.map_reqs_uids_to_paths.setdefault(
                    forward_requirement_.reserved_uid, OrderedSet()
                ).add(file_posix_path)

                if file_reference.g_file_entry.function is not None:
                    one_file_function_name_to_reqs_uids = (
                        self.map_file_function_names_to_reqs_uids.setdefault(
                            file_posix_path, {}
                        )
                    )
                    one_file_function_name_to_reqs_uids.setdefault(
                        file_reference.g_file_entry.function, []
                    ).append(
                        (forward_requirement_.reserved_uid, relation_.role)
                    )
                elif file_reference.g_file_entry.clazz is not None:
                    one_file_class_name_to_reqs_uids = (
                        self.map_file_class_names_to_reqs_uids.setdefault(
                            file_posix_path, {}
                        )
                    )
                    one_file_class_name_to_reqs_uids.setdefault(
                        file_reference.g_file_entry.clazz, []
                    ).append(
                        (forward_requirement_.reserved_uid, relation_.role)
                    )
                elif file_reference.g_file_entry.line_range is not None:
                    line_range = file_reference.g_file_entry.line_range
                    uid = forward_requirement_.reserved_uid
                    source_file_info = (
                        self.map_paths_to_source_file_traceability_info[
                            file_posix_path
                        ]
                    )
                    start_marker, end_marker = (
                        self.forward_range_markers_from_range(
                            line_range, uid, relation_.role
                        )
                    )
                    source_file_info.ng_map_reqs_to_markers.setdefault(
                        uid, []
                    ).append(start_marker)
                    source_file_info.markers.append(start_marker)
                    source_file_info.markers.append(end_marker)
                else:
                    uid = forward_requirement_.reserved_uid
                    source_file_info = (
                        self.map_paths_to_source_file_traceability_info[
                            file_posix_path
                        ]
                    )
                    forward_file_marker = (
                        self.forward_file_marker_from_file_info(
                            source_file_info,
                            uid,
                            relation_.role,
                        )
                    )
                    source_file_info.ng_map_reqs_to_markers.setdefault(
                        forward_requirement_.reserved_uid, []
                    ).append(forward_file_marker)
                    source_file_info.markers.append(forward_file_marker)

        #
        # STEP: Add markers for forward relations to functions and classes
        #
        for trace_info_ in self.trace_infos:
            source_file = assert_cast(trace_info_.source_file, SourceFile)

            self.map_paths_to_source_file_traceability_info[
                source_file.in_doctree_source_file_rel_path_posix
            ] = trace_info_

            for function_ in trace_info_.functions:
                # FIXME: Using display_name, not name. A separate exercise is
                #        to disambiguate forward links to C++ overloaded functions.
                if (
                    reqs_uids := self.get_req_uids_by_function_name(
                        source_file.in_doctree_source_file_rel_path_posix,
                        function_.display_name,
                    )
                ) is not None:
                    self.create_traceability_info_shared_markers_for_function(
                        trace_info_,
                        function_,
                        RangeMarkerType.FUNCTION,
                        reqs_uids,
                    )
                if (
                    reqs_uids := self.get_req_uids_by_class_name(
                        source_file.in_doctree_source_file_rel_path_posix,
                        function_.display_name,
                    )
                ) is not None:
                    self.create_traceability_info_shared_markers_for_function(
                        trace_info_,
                        function_,
                        RangeMarkerType.CLASS,
                        reqs_uids,
                    )

            validated_requirement_uids: Set[str] = set()
            marker_: Union[
                FunctionRangeMarker, LineMarker, RangeMarker, ForwardRangeMarker
            ]
            for marker_ in copy(trace_info_.markers):
                # FIXME: Is this 'continue' needed here?
                if isinstance(marker_, ForwardRangeMarker):
                    continue
                for requirement_uid_ in marker_.reqs:
                    if requirement_uid_ not in validated_requirement_uids:
                        node = traceability_index.get_node_by_uid_weak2(
                            requirement_uid_
                        )
                        if node is None:
                            raise StrictDocException(
                                f"Source file {source_file.in_doctree_source_file_rel_path_posix} references "
                                f"a requirement that does not exist: {requirement_uid_}."
                            )
                        validated_requirement_uids.add(requirement_uid_)

                    self.map_reqs_uids_to_paths.setdefault(
                        requirement_uid_, OrderedSet()
                    ).add(source_file.in_doctree_source_file_rel_path_posix)

                    node_id = traceability_index.get_node_by_uid(
                        requirement_uid_
                    )

                    self.map_paths_to_reqs.setdefault(
                        source_file.in_doctree_source_file_rel_path_posix,
                        OrderedSet(),
                    ).add(node_id)

                if isinstance(marker_, FunctionRangeMarker):
                    marker_copy = marker_.create_end_marker()
                    trace_info_.markers.append(marker_copy)

        #
        # Resolve definitions to declarations (only applicable for C and C++).
        #

        reversed_trace_info = {
            value: key
            for key, value in self.map_paths_to_source_file_traceability_info.items()
        }

        for (
            traceability_info_
        ) in self.map_paths_to_source_file_traceability_info.values():
            for function_ in traceability_info_.functions:
                if (
                    function_.is_declaration()
                    and function_.name
                    in self.map_all_function_names_to_definition_functions
                ):
                    definition_functions: List[Function] = []
                    if not function_.is_public():
                        definition_function = traceability_info_.ng_map_names_to_definition_functions.get(
                            function_.name, None
                        )
                        if definition_function is not None:
                            definition_functions.append(definition_function)
                    else:
                        mapped_definition_functions = (
                            self.map_all_function_names_to_definition_functions[
                                function_.name
                            ]
                        )
                        definition_functions.extend(mapped_definition_functions)
                    if len(definition_functions) == 0:
                        continue

                    for definition_function_ in definition_functions:
                        definition_function_trace_info: SourceFileTraceabilityInfo = definition_function_.parent

                        for marker_ in function_.markers:
                            function_marker = self.forward_function_marker_from_function(
                                function=definition_function_,
                                marker_type=RangeMarkerType.FUNCTION,
                                reqs=marker_.reqs_objs,
                                role=marker_.role,
                                description=f"function {function_.display_name}()",
                            )

                            for req_uid_ in marker_.reqs:
                                definition_function_trace_info.ng_map_reqs_to_markers.setdefault(
                                    req_uid_, []
                                ).append(function_marker)

                                path_to_info = reversed_trace_info[
                                    definition_function_trace_info
                                ]
                                self.map_reqs_uids_to_paths.setdefault(
                                    req_uid_, OrderedSet()
                                ).add(path_to_info)

                                node = traceability_index.get_node_by_uid(
                                    req_uid_
                                )
                                self.map_paths_to_reqs.setdefault(
                                    path_to_info, OrderedSet()
                                ).add(node)

                            definition_function_trace_info.markers.append(
                                function_marker
                            )

        #
        # STEP: For each trace info object:
        #       - Sort the markers according to their source location.
        #       - Calculate coverage information.
        #
        for (
            path,
            traceability_info_,
        ) in self.map_paths_to_source_file_traceability_info.items():

            def marker_comparator_start(
                marker: RelationMarkerType,
            ) -> int:
                assert marker.ng_range_line_begin is not None
                return marker.ng_range_line_begin

            sorted_markers = sorted(
                traceability_info_.markers, key=marker_comparator_start
            )

            traceability_info_.markers = sorted_markers
            # Finding how many lines are covered by the requirements in the file.
            # Quick and dirty: https://stackoverflow.com/a/15273749/598057
            merged_ranges: List[List[int]] = []
            for marker_ in traceability_info_.markers:
                assert isinstance(
                    marker_,
                    (
                        FunctionRangeMarker,
                        ForwardRangeMarker,
                        RangeMarker,
                        LineMarker,
                    ),
                ), marker_
                if marker_.ng_is_nodoc:
                    continue
                if not marker_.is_begin():
                    continue
                begin, end = (
                    assert_cast(marker_.ng_range_line_begin, int),
                    assert_cast(marker_.ng_range_line_end, int),
                )
                if merged_ranges and merged_ranges[-1][1] >= (begin - 1):
                    merged_ranges[-1][1] = max(merged_ranges[-1][1], end)
                else:
                    merged_ranges.append([begin, end])
            coverage = 0
            for merged_range in merged_ranges:
                for line_ in range(merged_range[0], merged_range[1] + 1):
                    if traceability_info_.file_stats.lines_info[line_]:
                        coverage += 1

            for function_ in traceability_info_.functions:
                for merged_range in merged_ranges:
                    if (
                        function_.line_begin >= merged_range[0]
                        and function_.line_end <= merged_range[1]
                    ):
                        traceability_info_.covered_functions += 1
                        break

            traceability_info_.set_coverage_stats(merged_ranges, coverage)

            for (
                req_uid_,
                markers_,
            ) in traceability_info_.ng_map_reqs_to_markers.items():

                def marker_comparator_range(
                    marker: RelationMarkerType,
                ) -> Tuple[int, int]:
                    assert marker.ng_range_line_begin is not None
                    assert marker.ng_range_line_end is not None
                    return marker.ng_range_line_begin, marker.ng_range_line_end

                markers_.sort(key=marker_comparator_range)

                # Validate here, SDocNode.relations doesn't track marker roles.
                node = traceability_index.get_node_by_uid(req_uid_)
                document = node.get_document()
                assert document is not None
                assert document.grammar is not None
                grammar_element = document.grammar.elements_by_type[
                    node.node_type
                ]
                for marker in markers_:
                    # Backwards markers do not require referenced node grammar
                    # to have the relation/role registered in the grammar.
                    if isinstance(marker, (FunctionRangeMarker, RangeMarker)):
                        continue

                    if not grammar_element.has_relation_type_role(
                        relation_type="File",
                        relation_role=marker.role,
                    ):
                        raise StrictDocSemanticError.invalid_marker_role(
                            node=node,
                            marker=marker,
                            path_to_src_file=path,
                        )

        # Sort by paths alphabetically.
        for paths_with_role in self.map_reqs_uids_to_paths.values():
            paths_with_role.sort()

        # Sort by node UID alphabetically.
        for path_requirements_ in self.map_paths_to_reqs.values():

            def compare_sdocnode_by_uid(node_: SDocNode) -> str:
                return assert_cast(node_.reserved_uid, str)

            path_requirements_.sort(key=compare_sdocnode_by_uid)

        #
        # STEP: Create auto-generated documents created from source file comments.
        #       Register these documents with the main traceability index.
        #
        def create_folder_section(
            parent__: Union[SDocDocumentIF, SDocNodeIF],
            document__: SDocDocument,
            path__: str,
        ) -> SDocNode:
            section_node = SDocNode(
                parent=parent__,
                node_type="SECTION",
                fields=[],
                relations=[],
                is_composite=True,
                node_type_close="SECTION",
            )
            section_node.ng_document_reference = DocumentReference()
            section_node.ng_document_reference.set_document(document__)
            section_node.ng_including_document_reference = DocumentReference()
            section_node.set_field_value(
                field_name="TITLE",
                form_field_index=0,
                value=path__,
            )
            return section_node

        section_cache = {}
        source_nodes_config: List[Dict[str, str]] = project_config.source_nodes
        for (
            path_to_source_file_,
            traceability_info_,
        ) in self.map_paths_to_source_file_traceability_info.items():
            if len(traceability_info_.source_nodes) == 0:
                continue

            if len(source_nodes_config) == 0:
                continue

            for config_entry_ in source_nodes_config:
                if path_to_source_file_.startswith(config_entry_["path"]):
                    relevant_source_node_entry = config_entry_
                    break
            else:
                continue

            document_uid = relevant_source_node_entry["uid"]
            document = traceability_index.get_node_by_uid(document_uid)

            current_top_node = document
            path_components = path_to_source_file_.split("/")
            for path_component_idx_, path_component_ in enumerate(
                path_components
            ):
                if path_component_ not in section_cache:
                    path_component_title = (
                        path_component_ + "/"
                        if path_component_idx_ < (len(path_components) - 1)
                        else path_component_
                    )
                    current_section = create_folder_section(
                        current_top_node, document, path_component_title
                    )
                    current_top_node.section_contents.append(current_section)
                    section_cache[path_component_] = current_section
                current_top_node = section_cache[path_component_]

            for source_node_ in traceability_info_.source_nodes:
                source_sdoc_node = SDocNode(
                    parent=document,
                    node_type=relevant_source_node_entry["node_type"],
                    fields=[],
                    relations=[],
                )
                assert source_node_.entity_name is not None
                source_sdoc_node_uid = f"{document_uid}/{path_to_source_file_}/{source_node_.entity_name}"
                source_sdoc_node.ng_document_reference = DocumentReference()
                source_sdoc_node.ng_document_reference.set_document(document)
                source_sdoc_node.ng_including_document_reference = (
                    DocumentReference()
                )

                source_sdoc_node.set_field_value(
                    field_name="UID",
                    form_field_index=0,
                    value=source_sdoc_node_uid,
                )
                source_sdoc_node.set_field_value(
                    field_name="TITLE",
                    form_field_index=0,
                    value=source_node_.entity_name,
                )

                for node_name_, node_value_ in source_node_.fields.items():
                    source_sdoc_node.set_field_value(
                        field_name=node_name_,
                        form_field_index=0,
                        value=node_value_,
                    )
                current_top_node.section_contents.append(source_sdoc_node)

                traceability_index.graph_database.create_link(
                    link_type=GraphLinkType.UID_TO_NODE,
                    lhs_node=source_sdoc_node_uid,
                    rhs_node=source_sdoc_node,
                )

                #
                # This connects:
                # - Source nodes and auto-generated requirements.
                # - Source nodes-related requirements and auto-generated requirements.
                #
                for marker_ in source_node_.markers:
                    if not isinstance(marker_, FunctionRangeMarker):
                        continue
                    traceability_info_.ng_map_reqs_to_markers.setdefault(
                        source_sdoc_node_uid, []
                    ).append(marker_)
                    self.map_reqs_uids_to_paths.setdefault(
                        source_sdoc_node_uid, OrderedSet()
                    ).add(path_to_source_file_)
                    self.map_paths_to_reqs.setdefault(
                        path_to_source_file_, OrderedSet()
                    ).add(source_sdoc_node)

                    for req_ in marker_.reqs:
                        node = traceability_index.get_node_by_uid_weak2(req_)
                        traceability_index.graph_database.create_link(
                            link_type=GraphLinkType.NODE_TO_PARENT_NODES,
                            lhs_node=source_sdoc_node,
                            rhs_node=node,
                        )
                        traceability_index.graph_database.create_link(
                            link_type=GraphLinkType.NODE_TO_CHILD_NODES,
                            lhs_node=node,
                            rhs_node=source_sdoc_node,
                        )

    def create_requirement_with_forward_source_links(
        self, requirement: SDocNode
    ) -> None:
        self.requirements_with_forward_links.add(requirement)

    def create_traceability_info(
        self,
        source_file: SourceFile,
        traceability_info: SourceFileTraceabilityInfo,
    ) -> None:
        assert isinstance(traceability_info, SourceFileTraceabilityInfo)
        traceability_info.source_file = source_file

        self.trace_infos.append(traceability_info)

    def get_req_uids_by_function_name(
        self, rel_path_posix: str, name: str
    ) -> Optional[List[Tuple[str, Optional[str]]]]:
        if rel_path_posix in self.map_file_function_names_to_reqs_uids:
            return self.map_file_function_names_to_reqs_uids[
                rel_path_posix
            ].get(name, None)
        return None

    def get_req_uids_by_class_name(
        self, rel_path_posix: str, name: str
    ) -> Optional[List[Tuple[str, Optional[str]]]]:
        if rel_path_posix in self.map_file_class_names_to_reqs_uids:
            return self.map_file_class_names_to_reqs_uids[rel_path_posix].get(
                name, None
            )
        return None

    @staticmethod
    def create_traceability_info_shared_markers_for_function(
        traceability_info: SourceFileTraceabilityInfo,
        function: Function,
        marker_type: RangeMarkerType,
        reqs_uids: List[Tuple[str, Optional[str]]],
    ) -> None:
        markers_by_role = {}
        for req_uid_, role in reqs_uids:
            req = Req(None, req_uid_)
            if role not in markers_by_role:
                markers_by_role[role] = (
                    FileTraceabilityIndex.forward_function_marker_from_function(
                        function, marker_type, [req], role
                    )
                )
            else:
                markers_by_role[role].reqs_objs.append(req)

        for req_uid_, role in reqs_uids:
            markers = traceability_info.ng_map_reqs_to_markers.setdefault(
                req_uid_, []
            )
            markers.append(markers_by_role[role])

        traceability_info.markers.extend(markers_by_role.values())

    @staticmethod
    def forward_function_marker_from_function(
        function: Function,
        marker_type: RangeMarkerType,
        reqs: List[Req],
        role: Optional[str],
        description: Optional[str] = None,
    ) -> ForwardFunctionRangeMarker:
        function_marker = ForwardFunctionRangeMarker(
            parent=None, reqs_objs=reqs, scope=marker_type.value
        )
        function_marker.ng_source_line_begin = function.line_begin
        function_marker.ng_range_line_begin = function.line_begin
        function_marker.ng_range_line_end = function.line_end
        function_marker.role = role
        if description is not None:
            function_marker.set_description(description)
        elif marker_type == RangeMarkerType.FUNCTION:
            function_marker.set_description(
                f"function {function.display_name}()"
            )
        elif marker_type == RangeMarkerType.CLASS:
            function_marker.set_description(f"class {function.name}")
        return function_marker

    @staticmethod
    def forward_range_markers_from_range(
        file_range: Tuple[int, int], requirement_uid_: str, role: Optional[str]
    ) -> Tuple[ForwardRangeMarker, ForwardRangeMarker]:
        start_marker = ForwardRangeMarker(
            start_or_end=True,
            reqs_objs=[Req(parent=None, uid=requirement_uid_)],
            role=role,
        )
        start_marker.ng_range_line_begin = file_range[0]
        start_marker.ng_source_line_begin = file_range[0]
        start_marker.ng_range_line_end = file_range[1]

        end_marker = ForwardRangeMarker(
            start_or_end=False,
            reqs_objs=[Req(parent=None, uid=requirement_uid_)],
            role=role,
        )
        end_marker.ng_source_line_begin = file_range[1]
        end_marker.ng_range_line_begin = file_range[0]
        end_marker.ng_range_line_end = file_range[1]

        return start_marker, end_marker

    @staticmethod
    def forward_file_marker_from_file_info(
        file_info: SourceFileTraceabilityInfo,
        requirement_uid_: str,
        role: Optional[str],
    ) -> ForwardFileMarker:
        marker = ForwardFileMarker(
            reqs_objs=[Req(parent=None, uid=requirement_uid_)],
            role=role,
        )
        marker.ng_range_line_begin = 1
        marker.ng_source_line_begin = 1
        marker.ng_range_line_end = file_info.file_stats.lines_total
        return marker



REGEX_UID = r"([\w]+[\w()\-\/.: ]*)"

NEGATIVE_MULTILINE_STRING_START = "(?!>>>\n)"
NEGATIVE_MULTILINE_STRING_END = "(?!^<<<)"
NEGATIVE_RELATIONS = "(?!^RELATIONS)"
NEGATIVE_UID = "(?!^UID)"
NEGATIVE_INLINE_LINK_START = rf"(?!\[LINK: {REGEX_UID})"
NEGATIVE_ANCHOR_START = rf"(?!^\[ANCHOR: {REGEX_UID})"

TEXT_TYPES_GRAMMAR = rf"""
TextPart[noskipws]:
  (Anchor | InlineLink | NormalString)
;

SingleLineTextPart[noskipws]:
  (Anchor | InlineLink | /{NEGATIVE_MULTILINE_STRING_START}\S.*/)
;

NormalString[noskipws]:
  (/(?ms){NEGATIVE_MULTILINE_STRING_END}{NEGATIVE_INLINE_LINK_START}{NEGATIVE_ANCHOR_START}./)+
;

InlineLink[noskipws]:
  '[LINK: ' value = /{REGEX_UID}/ ']'
;

Anchor[noskipws]:
  /^\[ANCHOR: /
  value = /{REGEX_UID}/ (', ' title = /\w+[\s\w+]*/)?
  /\](\Z|\n)/
;

// According to the Strict Grammar Rule #3, both SingleLineString and
// MultiLineString can never be empty strings.
// Both must eventualy start with a non-space character.
SingleLineString:
  /{NEGATIVE_MULTILINE_STRING_START}\S.*$/
;

MultiLineString[noskipws]:
  />>>\n/-
    parts*=TextPart
  /^<<</-
;

FieldName[noskipws]:
  /{NEGATIVE_UID}{NEGATIVE_RELATIONS}[A-Z]+[A-Z_0-9]*/
;
"""

FREE_TEXT_PARSER_GRAMMAR = r"""
FreeTextContainer[noskipws]:
  parts*=TextPart
;
"""

DOCUMENT_GRAMMAR = rf"""
SDocDocument[noskipws]:
  '[DOCUMENT]' '\n'
  ('MID: ' mid = SingleLineString '\n')?
  'TITLE: ' title = SingleLineString '\n'
  (config = DocumentConfig)?
  (view = DocumentView)?
  ('\n' grammar = DocumentGrammar)?
  section_contents *= SectionOrRequirement
;

DocumentConfig[noskipws]:
  ('UID: ' uid = /{REGEX_UID}/ '\n')?
  ('VERSION: ' version = SingleLineString '\n')?
  ('DATE: ' date = SingleLineString '\n')?
  ('CLASSIFICATION: ' classification = SingleLineString '\n')?
  (/(REQ_)?PREFIX/ ': ' requirement_prefix = SingleLineString '\n')?
  ('ROOT: ' (root = BooleanChoice) '\n')?
  ('OPTIONS:' '\n'
    ('  ENABLE_MID: ' (enable_mid = BooleanChoice) '\n')?
    ('  MARKUP: ' (markup = MarkupChoice) '\n')?
    ('  AUTO_LEVELS: ' (auto_levels = AutoLevelsChoice) '\n')?
    ('  LAYOUT: ' (layout = LayoutChoice) '\n')?
    ('  ' view_style_tag = /(VIEW_STYLE|REQUIREMENT_STYLE)/ ': '
        (requirement_style = RequirementStyleChoice) '\n')?
    ('  ' node_in_toc_tag = /(NODE_IN_TOC|REQUIREMENT_IN_TOC)/ ': '
        (requirement_in_toc = RequirementHasTitleChoice) '\n'
    )?
    ('  DEFAULT_VIEW: ' default_view = SingleLineString '\n')?
  )?
  ('METADATA:' '\n' custom_metadata = DocumentCustomMetadata)?
;

DocumentView[noskipws]:
  'VIEWS:' '\n'
  views += ViewElement
;

ViewElement[noskipws]:
  '- ID: ' view_id = /{REGEX_UID}/ '\n'
  ('  NAME: ' name = SingleLineString '\n')?
  '  TAGS:' '\n'
  tags += ViewElementTags
  ('  HIDDEN_TAGS:' '\n'
  hidden_tags += ViewElementHiddenTag)?
;

ViewElementTags[noskipws]:
  '  - OBJECT_TYPE: ' object_type = SingleLineString '\n'
  '    VISIBLE_FIELDS:' '\n'
  visible_fields += ViewElementField
;

ViewElementField[noskipws]:
  '    - NAME: ' name = SingleLineString '\n'
  ('      PLACEMENT: ' placement = SingleLineString '\n')?
;

ViewElementHiddenTag[noskipws]:
  '  - ' hidden_tag = SingleLineString '\n'
;

MarkupChoice[noskipws]:
  'RST' | 'Text' | 'HTML'
;

RequirementStyleChoice[noskipws]:
  'Plain' | 'Inline' | 'Simple' | 'Narrative' | 'Table' | 'Zebra'
;

RequirementHasTitleChoice[noskipws]:
  'True' | 'False'
;

AutoLevelsChoice[noskipws]:
  'On' | 'Off'
;

LayoutChoice[noskipws]:
  'Default' | 'Website'
;

DocumentCustomMetadata[noskipws]:
  (entries+=DocumentCustomMetadataKeyValuePair)*
;

DocumentCustomMetadataKeyValuePair[noskipws]:
  '  ' key=DocumentCustomMetadataKey ': ' value=SingleLineString '\n'
;

DocumentCustomMetadataKey: /[a-zA-Z_][a-zA-Z0-9_-]*/;

"""

SECTION_GRAMMAR = rf"""
SDocSection[noskipws]:
  '[SECTION]'
  '\n'
  ('MID: ' mid = SingleLineString '\n')?
  ('UID: ' uid = /{REGEX_UID}/ '\n')?
  ('LEVEL: ' custom_level = SingleLineString '\n')?
  'TITLE: ' title = SingleLineString '\n'
  (/(REQ_)?PREFIX/ ': ' requirement_prefix = SingleLineString '\n')?
  section_contents *= SectionOrRequirement
  '\n'
  '[/SECTION]'
  '\n'
;

SectionOrRequirement[noskipws]:
  '\n' (SDocSection | SDocCompositeNode | SDocNode | DocumentFromFile)
;

DocumentFromFile[noskipws]:
  '[DOCUMENT_FROM_FILE]' '\n'
  'FILE: ' file = /.+$/ '\n'
;

ReservedKeyword[noskipws]:
  'DOCUMENT' | 'GRAMMAR' | 'SECTION' | 'DOCUMENT_FROM_FILE'
;

SDocNode[noskipws]:
  '[' !'SECTION' node_type = RequirementType ']' '\n'
  fields += SDocNodeField
  (
    'RELATIONS:' '\n'
    (relations += Reference)
  )?
;

SDocCompositeNode[noskipws]:
  '[[' node_type = RequirementType ']]' '\n'

  fields += SDocNodeField
  (
    'RELATIONS:' '\n'
    (relations += Reference)
  )?

  section_contents *= SectionOrRequirement

  '\n'
  '[[/' node_type_close = RequirementType ']]' '\n'
;

SDocNodeField[noskipws]:
  (
    field_name = 'MID' ': ' parts+=SingleLineString '\n'
    |
    field_name = 'UID' ': ' parts+=/{REGEX_UID}/ '\n'
    |
    field_name = FieldName ':'
    (
        (' ' parts+=SingleLineTextPart '\n')
        |
        (
          ' '
          (
            multiline__ = />>>\n/
            parts+=TextPart
            /^<<</
          )
          '\n'
        )
    )
  )
;

RequirementStatus[noskipws]:
  'Draft' | 'Active' | 'Deleted';

"""



from typing import Any, Protocol

from textx import get_model


def drop_textx_meta(textx_object: Any) -> None:
    textx_object._tx_parser = None  # pylint: disable=protected-access
    textx_object._tx_attrs = None  # pylint: disable=protected-access
    textx_object._tx_metamodel = None  # pylint: disable=protected-access
    textx_object._tx_peg_rule = None  # pylint: disable=protected-access
    textx_object._tx_model_params = None  # pylint: disable=protected-access


def preserve_source_location_data(parsed_object: Any) -> None:
    """
    Saving the source location information in the parsed object.
    """
    the_model = get_model(parsed_object)
    line_start, col_start = the_model._tx_parser.pos_to_linecol(
        parsed_object._tx_position
    )
    line_end, col_end = the_model._tx_parser.pos_to_linecol(
        parsed_object._tx_position_end
    )
    parsed_object.ng_line_start = line_start
    parsed_object.ng_line_end = line_end
    parsed_object.ng_col_start = col_start
    parsed_object.ng_col_end = col_end
    parsed_object.ng_byte_start = parsed_object._tx_position
    parsed_object.ng_byte_end = parsed_object._tx_position_end


class SupportsTxPosition(Protocol):
    _tx_position: int



from dataclasses import dataclass
from typing import Dict, Iterator, List

from strictdoc.helpers.paths import SDocRelativePath


@dataclass
class AssetDir:
    full_path: str
    relative_path: SDocRelativePath

    def __post_init__(self) -> None:
        assert isinstance(self.relative_path, SDocRelativePath), (
            self.relative_path
        )


class AssetManager:
    def __init__(self) -> None:
        self.asset_dirs: List[AssetDir] = []
        self.asset_dirs_lookup: Dict[SDocRelativePath, AssetDir] = {}

    def add_asset_dir(
        self, full_path: str, relative_path: SDocRelativePath
    ) -> None:
        asset_dir = AssetDir(full_path=full_path, relative_path=relative_path)
        self.asset_dirs.append(asset_dir)

        self.asset_dirs_lookup[relative_path] = asset_dir

    def iterate(self) -> Iterator[AssetDir]:
        yield from self.asset_dirs



"""
Experimental code to generate SPDX.

This code is excluded from code coverage calculation because the feature is
highly experimental and most of this code will be removed.
"""

from typing import List, Union

from spdx_tools.spdx3.model import RelationshipType, SpdxDocument
from spdx_tools.spdx3.model.software import File, Package, Snippet

from strictdoc.backend.sdoc.document_reference import DocumentReference
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.document_grammar import (
    DocumentGrammar,
)
from strictdoc.backend.sdoc.models.grammar_element import (
    GrammarElement,
    GrammarElementFieldString,
    GrammarElementFieldType,
    GrammarElementRelationChild,
    GrammarElementRelationFile,
    GrammarElementRelationParent,
)
from strictdoc.backend.sdoc.models.model import (
    SDocDocumentIF,
    SDocNodeIF,
    SDocSectionIF,
)
from strictdoc.backend.sdoc.models.node import (
    SDocNode,
    SDocNodeField,
)
from strictdoc.backend.sdoc.models.reference import (
    ChildReqReference,
    FileEntry,
    FileReference,
    ParentReqReference,
)
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.export.spdx.spdx_sdoc_container import SPDXSDocContainer


class SPDXToSDocConverter:
    @staticmethod
    def convert(spdx_container: SPDXSDocContainer) -> SDocDocument:
        map_spdxref_to_sdoc = {}

        document = SDocDocument(
            mid=None,
            title=spdx_container.document.name,
            config=None,
            view=None,
            grammar=None,
            section_contents=[],
        )

        document.config.requirement_style = "Inline"

        document.grammar = SPDXToSDocConverter.create_grammar_for_spdx()
        document.grammar.parent = document

        #
        # Document.
        #
        document_requirement: SDocNode = SPDXToSDocConverter._convert_document(
            spdx_container.document,
            sdoc_document=document,
            sdoc_parent=document,
        )
        document.section_contents.append(document_requirement)

        #
        # Package.
        #
        package_requirement = SPDXToSDocConverter._convert_package(
            spdx_container.package,
            sdoc_document=document,
            sdoc_parent=document,
        )
        document.section_contents.append(package_requirement)

        #
        # Files.
        #

        file_section = SDocSection(
            parent=document,
            mid=None,
            uid=None,
            custom_level=None,
            title="Files",
            requirement_prefix=None,
            section_contents=[],
        )
        document.section_contents.append(file_section)

        for file_ in spdx_container.files:
            requirement = SPDXToSDocConverter._convert_file(
                file_, document, file_section
            )

            file_section.section_contents.append(requirement)
            map_spdxref_to_sdoc[file_.spdx_id] = requirement

        #
        # Snippets.
        #

        snippets_section = SDocSection(
            parent=document,
            mid=None,
            uid=None,
            custom_level=None,
            title="Snippets",
            requirement_prefix=None,
            section_contents=[],
        )
        document.section_contents.append(snippets_section)

        for snippet_ in spdx_container.snippets:
            requirement = SPDXToSDocConverter._convert_snippet(
                snippet_, document, file_section, spdx_container
            )

            snippets_section.section_contents.append(requirement)
            map_spdxref_to_sdoc[snippet_.spdx_id] = requirement

        for relationship_ in spdx_container.relationships:
            if (
                relationship_.from_element in map_spdxref_to_sdoc
                and relationship_.to[0] in map_spdxref_to_sdoc
            ):
                if relationship_.relationship_type == RelationshipType.CONTAINS:
                    from_element = spdx_container.map_spdx_ref_to_objects[
                        relationship_.from_element
                    ]
                    to_element = spdx_container.map_spdx_ref_to_objects[
                        relationship_.to[0]
                    ]

                    from_element_sdoc = map_spdxref_to_sdoc[
                        from_element.spdx_id
                    ]
                    to_element_sdoc = map_spdxref_to_sdoc[to_element.spdx_id]

                    assert to_element_sdoc.reserved_uid is not None, (
                        to_element_sdoc
                    )

                    from_element_sdoc.relations.append(
                        ChildReqReference(
                            parent=from_element_sdoc,
                            ref_uid=to_element_sdoc.reserved_uid,
                            role="CONTAINS",
                        )
                    )
                if (
                    relationship_.relationship_type
                    == RelationshipType.REQUIREMENT_FOR
                ):
                    from_element = spdx_container.map_spdx_ref_to_objects[
                        relationship_.from_element
                    ]
                    to_element = spdx_container.map_spdx_ref_to_objects[
                        relationship_.to[0]
                    ]

                    from_element_sdoc = map_spdxref_to_sdoc[
                        from_element.spdx_id
                    ]
                    to_element_sdoc = map_spdxref_to_sdoc[to_element.spdx_id]

                    assert to_element_sdoc.reserved_uid is not None, (
                        to_element_sdoc
                    )

                    from_element_sdoc.relations.append(
                        ParentReqReference(
                            parent=from_element_sdoc,
                            ref_uid=to_element_sdoc.reserved_uid,
                            role="REQUIREMENT_FOR",
                        )
                    )
        return document

    @staticmethod
    def _convert_document(
        document: SpdxDocument,
        sdoc_document: SDocDocument,
        sdoc_parent: Union[SDocDocumentIF, SDocSectionIF, SDocNodeIF],
    ) -> SDocNode:
        requirement = SDocNode(
            parent=sdoc_parent,
            node_type="SPDX_PACKAGE",
            fields=[],
            relations=[],
        )
        requirement.ng_document_reference = DocumentReference()
        requirement.ng_document_reference.set_document(sdoc_document)

        requirement.set_field_value(
            field_name="UID", form_field_index=0, value=document.spdx_id
        )
        requirement.set_field_value(
            field_name="SPDXID", form_field_index=0, value=document.spdx_id
        )
        requirement.set_field_value(
            field_name="TITLE", form_field_index=0, value=document.name
        )
        requirement.set_field_value(
            field_name="STATEMENT", form_field_index=0, value=document.summary
        )
        return requirement

    @staticmethod
    def _convert_package(
        package: Package,
        sdoc_document: SDocDocument,
        sdoc_parent: Union[SDocSection, SDocDocument],
    ) -> SDocNode:
        requirement = SDocNode(
            parent=sdoc_parent,
            node_type="SPDX_PACKAGE",
            fields=[],
            relations=[],
        )
        requirement.ng_document_reference = DocumentReference()
        requirement.ng_document_reference.set_document(sdoc_document)

        requirement.set_field_value(
            field_name="UID", form_field_index=0, value=package.spdx_id
        )
        requirement.set_field_value(
            field_name="SPDXID", form_field_index=0, value=package.spdx_id
        )
        requirement.set_field_value(
            field_name="PRIMARY_PURPOSE",
            form_field_index=0,
            value=package.primary_purpose.name,
        )
        requirement.set_field_value(
            field_name="TITLE", form_field_index=0, value=package.name
        )
        requirement.set_field_value(
            field_name="STATEMENT", form_field_index=0, value=package.summary
        )
        if package.summary is not None:
            requirement.set_field_value(
                field_name="SUMMARY", form_field_index=0, value=package.summary
            )
        return requirement

    @staticmethod
    def _convert_file(
        file: File,
        sdoc_document: SDocDocument,
        sdoc_parent: Union[SDocSection, SDocDocument],
    ) -> SDocNode:
        fields: List[SDocNodeField] = []
        requirement = SDocNode(
            parent=sdoc_parent,
            node_type="SPDX_FILE",
            fields=fields,
            relations=[],
        )
        requirement.ng_document_reference = DocumentReference()
        requirement.ng_document_reference.set_document(sdoc_document)

        requirement.set_field_value(
            field_name="UID", form_field_index=0, value=file.spdx_id
        )
        requirement.set_field_value(
            field_name="SPDXID", form_field_index=0, value=file.spdx_id
        )
        requirement.set_field_value(
            field_name="PRIMARY_PURPOSE",
            form_field_index=0,
            value=file.primary_purpose.name,
        )
        requirement.set_field_value(
            field_name="SUMMARY", form_field_index=0, value=file.summary
        )
        requirement.set_field_value(
            field_name="TITLE", form_field_index=0, value=file.name
        )

        requirement.relations = [
            FileReference(
                parent=requirement,
                g_file_entry=FileEntry(
                    parent=None,
                    g_file_format=None,
                    g_file_path=file.name,
                    g_line_range=None,
                    function=None,
                    clazz=None,
                ),
            )
        ]

        return requirement

    @staticmethod
    def _convert_snippet(
        snippet: Snippet,
        sdoc_document: SDocDocument,
        sdoc_parent: Union[SDocSection, SDocDocument],
        spdx_container: SPDXSDocContainer,
    ) -> SDocNode:
        fields: List[SDocNodeField] = []
        requirement = SDocNode(
            parent=sdoc_parent,
            node_type="SPDX_SNIPPET",
            fields=fields,
            relations=[],
        )
        requirement.ng_document_reference = DocumentReference()
        requirement.ng_document_reference.set_document(sdoc_document)

        requirement.set_field_value(
            field_name="UID", form_field_index=0, value=snippet.spdx_id
        )
        requirement.set_field_value(
            field_name="SPDXID", form_field_index=0, value=snippet.spdx_id
        )
        requirement.set_field_value(
            field_name="PRIMARY_PURPOSE",
            form_field_index=0,
            value=snippet.primary_purpose.name,
        )
        requirement.set_field_value(
            field_name="SUMMARY", form_field_index=0, value=snippet.summary
        )
        requirement.set_field_value(
            field_name="TITLE", form_field_index=0, value=snippet.name
        )

        spdx_file_id = spdx_container.map_spdx_snippets_to_files[
            snippet.spdx_id
        ]
        spdx_file = spdx_container.map_spdx_ref_to_objects[spdx_file_id]

        requirement.relations = [
            FileReference(
                parent=requirement,
                g_file_entry=FileEntry(
                    parent=None,
                    g_file_format=None,
                    g_file_path=spdx_file.name,
                    g_line_range=f"{snippet.line_range.begin}, {snippet.line_range.end - 1}",
                    function=None,
                    clazz=None,
                ),
            )
        ]

        return requirement

    @staticmethod
    def create_grammar_for_spdx() -> DocumentGrammar:
        elements = []

        #
        # SPDX Document
        #
        fields: List[GrammarElementFieldType] = [
            GrammarElementFieldString(
                parent=None,
                title="UID",
                human_title=None,
                required="True",
            ),
            GrammarElementFieldString(
                parent=None,
                title="SPDXID",
                human_title=None,
                required="True",
            ),
            GrammarElementFieldString(
                parent=None,
                title="TITLE",
                human_title=None,
                required="False",
            ),
            GrammarElementFieldString(
                parent=None,
                title="STATEMENT",
                human_title=None,
                required="False",
            ),
        ]

        document_element = GrammarElement(
            parent=None,
            tag="SPDX_DOCUMENT",
            property_is_composite="",
            property_prefix="",
            property_view_style="",
            fields=fields,
            relations=[
                GrammarElementRelationChild(
                    parent=None,
                    relation_type="Child",
                    relation_role="CONTAINS",
                ),
            ],
        )
        elements.append(document_element)

        #
        # SPDX Package
        #
        fields = [
            GrammarElementFieldString(
                parent=None,
                title="UID",
                human_title=None,
                required="True",
            ),
            GrammarElementFieldString(
                parent=None,
                title="SPDXID",
                human_title=None,
                required="True",
            ),
            GrammarElementFieldString(
                parent=None,
                title="PRIMARY_PURPOSE",
                human_title=None,
                required="False",
            ),
            GrammarElementFieldString(
                parent=None,
                title="TITLE",
                human_title=None,
                required="False",
            ),
            GrammarElementFieldString(
                parent=None,
                title="STATEMENT",
                human_title=None,
                required="False",
            ),
            GrammarElementFieldString(
                parent=None,
                title="SUMMARY",
                human_title=None,
                required="False",
            ),
        ]

        package_element = GrammarElement(
            parent=None,
            tag="SPDX_PACKAGE",
            property_is_composite="",
            property_prefix="",
            property_view_style="",
            fields=fields,
            relations=[
                GrammarElementRelationChild(
                    parent=None,
                    relation_type="Child",
                    relation_role="CONTAINS",
                ),
            ],
        )
        elements.append(package_element)

        #
        # SPDX File.
        #
        fields = [
            GrammarElementFieldString(
                parent=None,
                title="UID",
                human_title=None,
                required="True",
            ),
            GrammarElementFieldString(
                parent=None,
                title="SPDXID",
                human_title=None,
                required="True",
            ),
            GrammarElementFieldString(
                parent=None,
                title="PRIMARY_PURPOSE",
                human_title=None,
                required="False",
            ),
            GrammarElementFieldString(
                parent=None,
                title="SUMMARY",
                human_title=None,
                required="False",
            ),
            GrammarElementFieldString(
                parent=None,
                title="TITLE",
                human_title=None,
                required="False",
            ),
            GrammarElementFieldString(
                parent=None,
                title="STATEMENT",
                human_title=None,
                required="False",
            ),
        ]

        file_element = GrammarElement(
            parent=None,
            tag="SPDX_FILE",
            property_is_composite="",
            property_prefix="",
            property_view_style="",
            fields=fields,
            relations=[
                GrammarElementRelationChild(
                    parent=None,
                    relation_type="Child",
                    relation_role="CONTAINS",
                ),
                GrammarElementRelationFile(
                    parent=None,
                    relation_type="File",
                    relation_role=None,
                ),
            ],
        )
        elements.append(file_element)

        #
        # SPDX Snippet
        #
        fields = [
            GrammarElementFieldString(
                parent=None,
                title="UID",
                human_title=None,
                required="True",
            ),
            GrammarElementFieldString(
                parent=None,
                title="SPDXID",
                human_title=None,
                required="True",
            ),
            GrammarElementFieldString(
                parent=None,
                title="PRIMARY_PURPOSE",
                human_title=None,
                required="False",
            ),
            GrammarElementFieldString(
                parent=None,
                title="SUMMARY",
                human_title=None,
                required="False",
            ),
            GrammarElementFieldString(
                parent=None,
                title="TITLE",
                human_title=None,
                required="False",
            ),
            GrammarElementFieldString(
                parent=None,
                title="STATEMENT",
                human_title=None,
                required="False",
            ),
        ]

        elements.append(
            GrammarElement(
                parent=None,
                tag="SPDX_SNIPPET",
                property_is_composite="",
                property_prefix="",
                property_view_style="",
                fields=fields,
                relations=[
                    GrammarElementRelationParent(
                        parent=None,
                        relation_type="Parent",
                        relation_role="REQUIREMENT_FOR",
                    ),
                    GrammarElementRelationFile(
                        parent=None,
                        relation_type="File",
                        relation_role=None,
                    ),
                ],
            )
        )

        #
        # Create Grammar.
        #
        grammar = DocumentGrammar(parent=None, elements=elements)
        return grammar



import os
from dataclasses import dataclass
from typing import List, Optional

from strictdoc.cli.cli_arg_parser import ServerCommandConfig
from strictdoc.core.project_config import ProjectConfig


@dataclass
class UvicornReloadConfig:
    """
    Class that encapsulate the Uvicorn server configuration details.

    Uvicorn has an opinionated system for configuring the reloading of a server
    when the specified files are changed on the file system. This class
    centralized the reload configuration details in one place.

    One responsibility of this class is to give a correct configuration when
    the files must be reloaded and give a completely empty configuration when
    the reload is False. In particular, if a configuration specifies reload=False
    but the reload_dirs or reload_includes or reload_excludes are specified,
    then the following warning is shown.
    WARNING:  Current configuration will not reload as not all conditions are met, please refer to documentation.
    The create() method of this class ensures that uvicorn is happy and the
    warning does not appear.
    """

    reload: bool
    reload_dirs: Optional[List[str]]
    reload_includes: Optional[List[str]]
    reload_excludes: Optional[List[str]]

    @classmethod
    def create(
        cls, project_config: ProjectConfig, server_config: ServerCommandConfig
    ) -> "UvicornReloadConfig":
        reload: bool = server_config.reload
        reload_dirs: Optional[List[str]] = None
        reload_includes: Optional[List[str]] = None
        reload_excludes: Optional[List[str]] = None

        if server_config.reload:
            # This doesn't seem to be effect because we still must provide the
            # reload_excludes outside this folder. Maybe it has to do with the
            # presence of reload_includes that includes all files with given
            # file extensions below. Anyway this works well enough with the
            # given reload_excludes.
            reload_dirs = [
                os.path.join(
                    project_config.get_strictdoc_root_path(), "strictdoc"
                )
            ]
            # It is important that StrictDoc's artifacts are excluded because
            # StrictDoc's web server will be triggered to restart every time the
            # contents of these folders changes.
            reload_excludes = (
                cls.expand_folder("build", max_depth=15)
                + cls.expand_folder("docs", max_depth=3)
                + cls.expand_folder("dist", max_depth=3)
                + cls.expand_folder("tests", max_depth=15)
                + cls.expand_folder("output", max_depth=15)
                + cls.expand_folder("Output", max_depth=15)
            )

            # Changing typical StrictDoc's own source code files should trigger
            # restarting of the web server. Whitelisting them here.
            reload_includes = [
                "*.py",
                "*.html",
                "*.jinja",
                "*.css",
                "*.js",
                "*.toml",
            ]
        return UvicornReloadConfig(
            reload=reload,
            reload_dirs=reload_dirs,
            reload_includes=reload_includes,
            reload_excludes=reload_excludes,
        )

    @classmethod
    def expand_folder(cls, folder: str, max_depth: int) -> List[str]:
        """
        Create a list of wildcard-based paths for a given folder.

        It looks like the regex engine of the uvicorn file watching library
        does not support proper ** globs.

        Examples:
        "tests",  # Doesn't work.
        "tests/",  # Doesn't work.
        "tests/*",  # Doesn't work.
        "tests/**",  # Makes the process hang, server doesn't start.

        Example of a possible StrictDoc path that must be excluded:
        output/cache/server/html/_source_files/strictdoc/backend/reqif/sdoc_reqif_fields.py.html

        This function produces output like this:
        [
            "output/*.*",
            "output/*/*.*",
            "output/*/*/*.*",
            "output/*/*/*/*.*",
            "output/*/*/*/*/*.*",
            "output/*/*/*/*/*/*.*",
            "output/*/*/*/*/*/*/*.*",
            "output/*/*/*/*/*/*/*/*.*",
            "output/*/*/*/*/*/*/*/*/*.*",
            "output/*/*/*/*/*/*/*/*/*/*.*",
            "output/*/*/*/*/*/*/*/*/*/*/*.*",
            "output/*/*/*/*/*/*/*/*/*/*/*/*.*",
            "output/*/*/*/*/*/*/*/*/*/*/*/*/*.*",
            "output/*/*/*/*/*/*/*/*/*/*/*/*/*/*.*",
            "output/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*.*",
        ]
        """

        patterns = []
        for depth in range(1, max_depth + 1):
            pattern = folder + "/" + "*/" * (depth - 1) + "*.*"
            patterns.append(pattern)
        return patterns



import os.path
from typing import Callable, Dict, List, Optional, Union, cast

from textx import TextXSyntaxError, get_model

from strictdoc.backend.sdoc.document_reference import DocumentReference
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.document_config import DocumentConfig
from strictdoc.backend.sdoc.models.document_from_file import DocumentFromFile
from strictdoc.backend.sdoc.models.document_grammar import DocumentGrammar
from strictdoc.backend.sdoc.models.document_view import DocumentView
from strictdoc.backend.sdoc.models.grammar_element import GrammarElement
from strictdoc.backend.sdoc.models.model import (
    SDocDocumentFromFileIF,
    SDocDocumentIF,
    SDocNodeIF,
    SDocSectionIF,
)
from strictdoc.backend.sdoc.models.node import (
    SDocNode,
    SDocNodeField,
)
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.helpers.exception import StrictDocException
from strictdoc.helpers.textx import (
    SupportsTxPosition,
    preserve_source_location_data,
)


class ParseContext:
    def __init__(
        self, path_to_sdoc_file: Optional[str], migrate_sections: bool = False
    ):
        self.path_to_sdoc_file: Optional[str] = path_to_sdoc_file
        self.path_to_sdoc_dir: Optional[str] = None
        if path_to_sdoc_file is not None:
            self.path_to_sdoc_dir = os.path.dirname(path_to_sdoc_file)
        self.document_grammar: Optional[DocumentGrammar] = None
        self.document_reference: DocumentReference = DocumentReference()
        self.context_document_reference: DocumentReference = DocumentReference()
        self.document_config: Optional[DocumentConfig] = None
        self.document_view: Optional[DocumentView] = None
        self.document_has_requirements = False

        self.fragments_from_files: List[SDocDocumentFromFileIF] = []
        self.migrate_sections: bool = migrate_sections


class SDocParsingProcessor:
    def __init__(self, parse_context: ParseContext) -> None:
        self.parse_context: ParseContext = parse_context

    def process_document(self, document: SDocDocument) -> None:
        document.grammar = (
            self.parse_context.document_grammar
            or DocumentGrammar.create_default(
                document,
                create_section_element=self.parse_context.migrate_sections,
                enable_mid=document.config.enable_mid or False,
            )
        )
        document.ng_including_document_reference = (
            self.parse_context.context_document_reference
        )

    def get_default_processors(self) -> Dict[str, Callable[..., None]]:
        return {
            "SDocDocument": self.process_document,
            "DocumentConfig": self.process_document_config,
            "DocumentGrammar": self.process_document_grammar,
            "GrammarElement": self.process_document_grammar_element,
            "DocumentView": self.process_document_view,
            "SDocSection": self.process_section,
            "DocumentFromFile": self.process_document_from_file,
            "SDocCompositeNode": self.process_requirement,
            "SDocNode": self.process_requirement,
            "SDocNodeField": self.process_node_field,
        }

    def process_document_config(self, document_config: DocumentConfig) -> None:
        the_model = get_model(document_config)
        line_start, col_start = the_model._tx_parser.pos_to_linecol(
            cast(SupportsTxPosition, document_config)._tx_position
        )
        document_config.ng_line_start = line_start
        document_config.ng_col_start = col_start
        self.parse_context.document_config = document_config

    def process_document_grammar(
        self, document_grammar: DocumentGrammar
    ) -> None:
        assert self.parse_context.document_config is not None

        preserve_source_location_data(document_grammar)
        # FIXME: It would be great to move forward and remove this.
        if not document_grammar.has_text_element():
            document_grammar.add_element_first(
                DocumentGrammar.create_default_text_element(
                    document_grammar,
                    enable_mid=self.parse_context.document_config.enable_mid
                    is True,
                )
            )
        self.parse_context.document_grammar = document_grammar

    def process_document_grammar_element(
        self, grammar_element: GrammarElement
    ) -> None:
        preserve_source_location_data(grammar_element)

    def process_document_view(self, document_view: DocumentView) -> None:
        self.parse_context.document_view = document_view

        the_model = get_model(document_view)
        line_start, col_start = the_model._tx_parser.pos_to_linecol(
            cast(SupportsTxPosition, document_view)._tx_position
        )
        document_view.ng_line_start = line_start
        document_view.ng_col_start = col_start

    def process_section(self, section: SDocSection) -> None:
        section.ng_document_reference = self.parse_context.document_reference
        section.ng_including_document_reference = (
            self.parse_context.context_document_reference
        )
        preserve_source_location_data(section)

        # FIXME: Refactor to eliminate the need in such assert.
        assert self.parse_context.document_config is not None

        if self.parse_context.document_config.auto_levels:
            if (
                section.ng_resolved_custom_level
                and section.ng_resolved_custom_level != "None"
            ):
                raise StrictDocException(
                    "[SECTION].LEVEL field is provided. "
                    "This contradicts to the option "
                    "[DOCUMENT].OPTIONS.AUTO_LEVELS set to On. "
                    f"Section: {section}"
                )
        else:
            if not section.ng_resolved_custom_level:
                raise StrictDocException(
                    "[SECTION].LEVEL field is not provided. "
                    "This contradicts to the option "
                    "[DOCUMENT].OPTIONS.AUTO_LEVELS set to Off. "
                    f"Section: {section}"
                )

    def process_document_from_file(
        self, document_from_file: DocumentFromFile
    ) -> None:
        assert isinstance(document_from_file, DocumentFromFile), (
            document_from_file
        )

        # Windows paths are backslashes, so using abspath in addition.
        assert self.parse_context.path_to_sdoc_dir is not None
        resolved_path_to_fragment_file = os.path.abspath(
            os.path.join(
                self.parse_context.path_to_sdoc_dir, document_from_file.file
            )
        )
        if not resolved_path_to_fragment_file.endswith(".sdoc"):
            raise StrictDocException(
                '[DOCUMENT_FROM_FILE]: A document file name must have ".sdoc" extension: '
                f"{document_from_file.file}."
            )
        if not os.path.isfile(resolved_path_to_fragment_file):
            raise StrictDocException(
                "[DOCUMENT_FROM_FILE]: Path to a document file does not exist: "
                f"{document_from_file.file}."
            )

        document_from_file.ng_document_reference = (
            self.parse_context.document_reference
        )
        document_from_file.resolved_full_path_to_document_file = (
            resolved_path_to_fragment_file
        )

        self.parse_context.fragments_from_files.append(document_from_file)

    def process_requirement(self, requirement: SDocNode) -> None:
        self.parse_context.document_has_requirements = True

        requirement.ng_document_reference = (
            self.parse_context.document_reference
        )
        requirement.ng_including_document_reference = (
            self.parse_context.context_document_reference
        )

        cursor: Union[SDocDocumentIF, SDocSectionIF, SDocNodeIF] = (
            requirement.parent
        )
        while (
            isinstance(cursor, (SDocSectionIF, SDocNodeIF))
            and not cursor.ng_has_requirements
        ):
            cursor.ng_has_requirements = True
            cursor = cursor.parent

        assert self.parse_context.document_config is not None
        if (
            requirement.reserved_title is None
            or not self.parse_context.document_config.is_requirement_in_toc()
        ) and self.parse_context.document_config.auto_levels:
            requirement.ng_resolved_custom_level = "None"

        preserve_source_location_data(requirement)

    def process_node_field(self, node_field: SDocNodeField) -> None:
        node_field_parts = node_field.parts
        if (
            isinstance(node_field_parts[0], str)
            and node_field_parts[0].strip() == ""
        ):
            the_model = get_model(node_field)
            line_start, col_start = the_model._tx_parser.pos_to_linecol(
                cast(SupportsTxPosition, node_field)._tx_position
            )
            raise TextXSyntaxError(
                "Node statement cannot be empty.",
                line=line_start,
                col=col_start,
                filename=self.parse_context.path_to_sdoc_file,
            )



from typing import Dict, List, Optional, Set

from strictdoc.backend.sdoc.models.grammar_element import (
    GrammarElement,
    GrammarElementFieldString,
    GrammarElementFieldType,
)
from strictdoc.backend.sdoc.models.model import (
    RequirementFieldName,
    SDocDocumentIF,
    SDocGrammarIF,
)


class DocumentGrammar(SDocGrammarIF):
    def __init__(
        self,
        parent: Optional[SDocDocumentIF],
        elements: List[GrammarElement],
        import_from_file: Optional[str] = None,
    ) -> None:
        self.parent: Optional[SDocDocumentIF] = parent
        self.elements: List[GrammarElement] = elements

        self.registered_elements: Set[str] = set()
        self.elements_by_type: Dict[str, GrammarElement] = {}

        self.update_with_elements(elements)

        # The textX parser passes an empty string instead of None when
        # import_from_file is not provided in input.
        if import_from_file is not None and len(import_from_file) == 0:
            import_from_file = None
        self.import_from_file: Optional[str] = import_from_file

        self.is_default = False

        self.ng_line_start: Optional[int] = None
        self.ng_col_start: Optional[int] = None

    @staticmethod
    def create_default(
        parent: Optional[SDocDocumentIF],
        create_section_element: bool = False,
        enable_mid: bool = False,
    ) -> "DocumentGrammar":
        text_element: GrammarElement = (
            DocumentGrammar.create_default_text_element(enable_mid=enable_mid)
        )

        # @relation(SDOC-SRS-132, scope=range_start)
        fields: List[GrammarElementFieldType] = []

        if enable_mid:
            fields.append(
                GrammarElementFieldString(
                    parent=None,
                    title=RequirementFieldName.MID,
                    human_title=None,
                    required="False",
                )
            )

        fields.extend(
            [
                GrammarElementFieldString(
                    parent=None,
                    title=RequirementFieldName.UID,
                    human_title=None,
                    required="False",
                ),
                GrammarElementFieldString(
                    parent=None,
                    title=RequirementFieldName.LEVEL,
                    human_title=None,
                    required="False",
                ),
                GrammarElementFieldString(
                    parent=None,
                    title=RequirementFieldName.STATUS,
                    human_title=None,
                    required="False",
                ),
                GrammarElementFieldString(
                    parent=None,
                    title=RequirementFieldName.TAGS,
                    human_title=None,
                    required="False",
                ),
                GrammarElementFieldString(
                    parent=None,
                    title=RequirementFieldName.TITLE,
                    human_title=None,
                    required="False",
                ),
                GrammarElementFieldString(
                    parent=None,
                    title=RequirementFieldName.STATEMENT,
                    human_title=None,
                    required="False",
                ),
                GrammarElementFieldString(
                    parent=None,
                    title=RequirementFieldName.RATIONALE,
                    human_title=None,
                    required="False",
                ),
                GrammarElementFieldString(
                    parent=None,
                    title=RequirementFieldName.COMMENT,
                    human_title=None,
                    required="False",
                ),
            ]
        )

        requirement_element = GrammarElement(
            parent=None,
            tag="REQUIREMENT",
            property_is_composite="",
            property_prefix="",
            property_view_style="",
            fields=fields,
            relations=[],
        )
        # @relation(SDOC-SRS-132, scope=range_end)

        requirement_element.relations = GrammarElement.create_default_relations(
            requirement_element
        )

        elements: List[GrammarElement] = []

        if create_section_element:
            section_element: GrammarElement = (
                DocumentGrammar.create_default_section_element(
                    enable_mid=enable_mid
                )
            )
            elements.append(section_element)

        elements.append(text_element)
        elements.append(requirement_element)

        grammar = DocumentGrammar(
            parent=parent, elements=elements, import_from_file=None
        )
        grammar.is_default = True
        text_element.parent = grammar
        requirement_element.parent = grammar

        return grammar

    @staticmethod
    def create_for_test_report(parent: SDocDocumentIF) -> "DocumentGrammar":
        text_element: GrammarElement = (
            DocumentGrammar.create_default_text_element()
        )

        fields: List[GrammarElementFieldType] = [
            GrammarElementFieldString(
                parent=None,
                title=RequirementFieldName.UID,
                human_title=None,
                required="False",
            ),
            GrammarElementFieldString(
                parent=None,
                title="TEST_PATH",
                human_title=None,
                required="False",
            ),
            GrammarElementFieldString(
                parent=None,
                title="TEST_FUNCTION",
                human_title=None,
                required="False",
            ),
            GrammarElementFieldString(
                parent=None,
                title="DURATION",
                human_title=None,
                required="False",
            ),
            GrammarElementFieldString(
                parent=None,
                title=RequirementFieldName.STATUS,
                human_title=None,
                required="True",
            ),
            GrammarElementFieldString(
                parent=None,
                title=RequirementFieldName.TITLE,
                human_title=None,
                required="True",
            ),
            GrammarElementFieldString(
                parent=None,
                title=RequirementFieldName.STATEMENT,
                human_title=None,
                required="False",
            ),
        ]
        requirement_element = GrammarElement(
            parent=None,
            tag="TEST_RESULT",
            property_is_composite="",
            property_prefix="",
            property_view_style="",
            fields=fields,
            relations=[],
        )

        requirement_element.relations = GrammarElement.create_default_relations(
            requirement_element
        )

        elements: List[GrammarElement] = [text_element, requirement_element]
        grammar = DocumentGrammar(
            parent=parent, elements=elements, import_from_file=None
        )
        grammar.is_default = True
        text_element.parent = grammar
        requirement_element.parent = grammar

        return grammar

    def get_element_by_mid(self, element_mid: str) -> GrammarElement:
        for element_ in self.elements:
            if element_.mid == element_mid:
                return element_
        raise AssertionError(
            f"Could not find a grammar element with MID: {element_mid}"
        )

    def dump_fields(self, node_type: str) -> str:
        return ", ".join(
            list(
                map(
                    lambda g: g.title,
                    self.elements_by_type[node_type].fields,
                )
            )
        )

    def has_text_element(self) -> bool:
        for element_ in self.elements:
            if element_.tag == "TEXT":
                return True
        return False

    def add_element_first(self, element: GrammarElement) -> None:
        self.elements.insert(0, element)
        self.elements_by_type[element.tag] = element
        self.registered_elements.add(element.tag)
        self.is_default = False

    def update_element(
        self, existing_element: GrammarElement, updated_element: GrammarElement
    ) -> None:
        element_index = self.elements.index(existing_element)
        self.elements[element_index] = updated_element
        self.elements_by_type[updated_element.tag] = updated_element
        self.is_default = False

    def update_with_elements(self, elements: List[GrammarElement]) -> None:
        # When elements are created by code, not by textX, it is convenient
        # if their .parent is set here automatically.
        for element_ in elements:
            element_.parent = self

        registered_elements: Set[str] = set()
        elements_by_type: Dict[str, GrammarElement] = {}
        for element in elements:
            registered_elements.add(element.tag)
            elements_by_type[element.tag] = element

        self.elements = elements
        self.registered_elements = registered_elements
        self.elements_by_type = elements_by_type

    @staticmethod
    def create_default_text_element(
        parent: Optional["DocumentGrammar"] = None,
        enable_mid: bool = False,
    ) -> GrammarElement:
        fields: List[GrammarElementFieldType] = []
        if enable_mid:
            fields.append(
                GrammarElementFieldString(
                    parent=None,
                    title=RequirementFieldName.MID,
                    human_title=None,
                    required="False",
                )
            )
        fields.extend(
            [
                GrammarElementFieldString(
                    parent=None,
                    title=RequirementFieldName.UID,
                    human_title=None,
                    required="False",
                ),
                GrammarElementFieldString(
                    parent=None,
                    title=RequirementFieldName.STATEMENT,
                    human_title=None,
                    required="True",
                ),
            ]
        )
        text_element = GrammarElement(
            parent=parent,
            tag="TEXT",
            property_is_composite="",
            property_prefix="",
            property_view_style="",
            fields=fields,
            relations=[],
        )
        return text_element

    @staticmethod
    def create_default_section_element(
        parent: Optional["DocumentGrammar"] = None, enable_mid: bool = False
    ) -> GrammarElement:
        fields: List[GrammarElementFieldType] = []
        if enable_mid:
            fields.append(
                GrammarElementFieldString(
                    parent=None,
                    title=RequirementFieldName.MID,
                    human_title=None,
                    required="True",
                )
            )
        fields.append(
            GrammarElementFieldString(
                parent=None,
                title=RequirementFieldName.UID,
                human_title=None,
                required="False",
            )
        )
        fields.append(
            GrammarElementFieldString(
                parent=None,
                title=RequirementFieldName.LEVEL,
                human_title=None,
                required="False",
            )
        )
        fields.append(
            GrammarElementFieldString(
                parent=None,
                title=RequirementFieldName.PREFIX,
                human_title=None,
                required="False",
            )
        )
        fields.append(
            GrammarElementFieldString(
                parent=None,
                title=RequirementFieldName.TITLE,
                human_title=None,
                required="True",
            ),
        )
        section_element = GrammarElement(
            parent=parent,
            tag="SECTION",
            property_is_composite="True",
            property_prefix="",
            property_view_style="",
            fields=fields,
            relations=[],
        )
        return section_element


class DocumentGrammarWrapper:
    def __init__(self, grammar: DocumentGrammar):
        assert isinstance(grammar, DocumentGrammar), grammar
        self.grammar: DocumentGrammar = grammar
